"""Internal module for additional GIS database functions."""

from __future__ import annotations

import logging
from functools import lru_cache, reduce

from django.contrib.gis.db.models import Extent, PolygonField, functions
from django.contrib.gis.db.models.fields import ExtentField
from django.db import connection, connections, models

from gisserver import conf
from gisserver.crs import CRS, WGS84
from gisserver.geometries import WGS84BoundingBox
from gisserver.types import GeometryXsdElement

logger = logging.getLogger(__name__)


class AsEWKT(functions.GeoFunc):
    """Generate EWKT in the database (PostGIS tested only at the moment)."""

    name = "AsEWKT"
    output_field = models.TextField()

    def __init__(self, field, precision=conf.GISSERVER_DB_PRECISION, **extra):
        if connection.ops.spatial_version >= (3, 1):
            super().__init__(field, precision, **extra)
        else:
            super().__init__(field, **extra)


class AsGML(functions.AsGML):
    """An overwritten ST_AsGML() function to handle PostGIS extensions."""

    name = "AsGML"

    def __init__(
        self,
        expression,
        version=3,
        precision=conf.GISSERVER_DB_PRECISION,
        envelope=False,
        is_latlon=False,
        long_urn=False,
        **extra,
    ):
        # Note that Django's AsGml the defaults are: version=2, precision=8
        super().__init__(expression, version, precision, **extra)
        self.envelope = envelope
        self.is_latlon = is_latlon
        self.long_urn = long_urn

    def as_postgresql(self, compiler, connection, **extra_context):
        # Fill options parameter (https://postgis.net/docs/ST_AsGML.html)
        options = 0
        if self.long_urn:
            options |= 1  # long CRS urn
        if self.envelope:
            options |= 32  # bbox
        if self.is_latlon:
            # PostGIS provides the data in longitude/latitude format (east/north to look like x/y).
            # However, WFS 2.0 fixed their axis by following the authority. The ST_AsGML() doesn't
            # detect this on their own. It needs to be told to flip the coordinates.
            # Passing option 16 flips the coordinates unconditionally, like using ST_FlipCoordinates()
            # but more efficiently done during rendering. That happens in:
            # https://github.com/postgis/postgis/blob/81e2bc783b77cc740291445e992658e1db7179e0/liblwgeom/lwout_gml.c#L121
            options |= 16

        template = f"%(function)s(%(expressions)s, {options})"
        return self.as_sql(compiler, connection, template=template, **extra_context)


class ST_SetSRID(functions.Transform):
    """PostGIS function to assign an SRID to geometry.
    When this is applied to the result from an ``Extent`` aggegrate,
    it will convert that ``BBOX(...)`` value into a ``POLYGON(...)``.
    """

    name = "SetSRID"
    geom_param_pos = ()

    @property
    def geo_field(self):
        return PolygonField(srid=self.source_expressions[1].value)


class Box2D(functions.GeomOutputGeoFunc):
    """PostGIS function (without ST_) that converts a ``POLYGON(...)`` back to a ``BOX(...)``."""

    name = "Box2D"
    function = "Box2D"  # no ST_ prefix.
    output_field = ExtentField()

    def convert_value(self, value, expression, connection):
        return connection.ops.convert_extent(value)


class ST_Union(functions.Union):
    name = "Union"
    arity = None

    def as_postgresql(self, compiler, connection, **extra_context):
        # PostgreSQL can handle ST_Union(ARRAY[field names]), other databases don't.
        if len(self.source_expressions) > 2:
            extra_context["template"] = "%(function)s(ARRAY[%(expressions)s])"
        return self.as_sql(compiler, connection, **extra_context)


def get_wgs84_bounding_box(
    queryset: models.QuerySet, geo_element: GeometryXsdElement
) -> WGS84BoundingBox:
    """Calculate the WGS84 bounding box for a feature.

    Note that the ``<ows:WGS84BoundingBox>`` element
    always uses longitude/latitude, and doesn't describe a CRS.
    """
    if connections[queryset.db].vendor == "postgresql":
        # Allow a more efficient way to combine geometry first, transform once later
        box = queryset.aggregate(
            box=Box2D(
                functions.Transform(
                    ST_SetSRID(Extent(geo_element.orm_path), srid=geo_element.source_srid),
                    srid=WGS84.srid,
                )
            )
        )["box"]
    else:
        # Need to transform each element to srid 4326 before combining it in an extent.
        box = queryset.aggregate(box=Extent(get_db_geometry_target(geo_element, WGS84)))["box"]

    return WGS84BoundingBox(*box) if box else None


def get_geometries_union(
    expressions: list[str | functions.GeoFunc], using="default"
) -> str | functions.Union:
    """Generate a union of multiple geometry fields."""
    if not expressions:
        raise ValueError("Missing geometry fields for get_geometries_union()")

    if len(expressions) == 1:
        return next(iter(expressions))  # fastest in set data type
    elif len(expressions) == 2:
        return functions.Union(*expressions)
    elif connections[using].vendor == "postgresql":
        # postgres can handle multiple field names
        return ST_Union(*expressions)
    else:
        # other databases do Union(Union(1, 2), 3)
        return reduce(functions.Union, expressions)


def replace_queryset_geometries(
    queryset: models.QuerySet,
    geo_elements: list[GeometryXsdElement],
    output_crs: CRS,
    wrapper_func: type[functions.GeoFunc],
    **wrapper_kwargs,
) -> models.QuerySet:
    """Replace the queryset geometry retrieval with a database-rendered version.

    This uses absolute paths in the queryset, but can use relative paths for related querysets.
    """
    defer_names = []
    as_geo_map = {}
    for geo_element in geo_elements:
        if geo_element.source is not None:  # excludes GmlBoundedByElement
            defer_names.append(geo_element.local_orm_path)
            annotation_name = _as_annotation_name(geo_element.local_orm_path, wrapper_func)
            as_geo_map[annotation_name] = wrapper_func(
                get_db_geometry_target(geo_element, output_crs, use_relative_path=True),
                **wrapper_kwargs,
            )

    if not defer_names:
        return queryset

    logger.debug(
        "DB rendering: QuerySet for %s replacing %r with %r",
        queryset.model._meta.label,
        defer_names,
        list(as_geo_map.keys()),
    )
    return queryset.defer(*defer_names).annotate(**as_geo_map)


def get_db_rendered_geometry(
    instance: models.Model, geo_element: GeometryXsdElement, replacement: type[functions.GeoFunc]
) -> str:
    """Retrieve the database-rendered geometry.
    This includes formatted EWKT or GML output, rendered by the database.
    """
    annotation_name = _as_annotation_name(geo_element.local_orm_path, replacement)
    try:
        return getattr(instance, annotation_name)
    except AttributeError as e:
        prefix = _as_annotation_name("", replacement)
        available = ", ".join(key for key in instance.__dict__ if key.startswith(prefix)) or "none"
        raise AttributeError(
            f" DB annotation {instance._meta.label}.{annotation_name} not found. Found {available}."
        ) from e


@lru_cache
def _as_annotation_name(name: str, func: type[functions.GeoFunc]) -> str:
    """Escape an XML name to be used as annotation name."""
    return f"_{func.__name__}_{name.replace('.', '_')}"


def get_db_geometry_target(
    geo_element: GeometryXsdElement, output_crs: CRS, use_relative_path: bool = False
) -> str | functions.Transform:
    """Translate a GML geometry field into the proper expression for retrieving it from the database.
    The path will be wrapped into a CRS Transform function if needed.
    """
    orm_path = geo_element.local_orm_path if use_relative_path else geo_element.orm_path
    if geo_element.source_srid != output_crs.srid:
        return functions.Transform(orm_path, srid=output_crs.srid)
    else:
        return orm_path



"""WFS 2.0 request objects for parsing.

These objects convert the various request formats into a uniform request object.
The object properties are based on the WFS specification and XSD type names.
By following these definitions closely, it naturally follows to support nearly
all possible request formats outside the common examples.

Examples:

* https://schemas.opengis.net/wfs/2.0/examples/
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from gisserver.exceptions import (
    InvalidParameterValue,
    MissingParameterValue,
    OperationParsingFailed,
)
from gisserver.parsers import fes20
from gisserver.parsers.ast import expect_no_children, tag_registry
from gisserver.parsers.ows import BaseOwsRequest, KVPRequest
from gisserver.parsers.xml import NSElement, xmlns

from .adhoc import AdhocQuery
from .base import QueryExpression
from .projection import ResolveValue, parse_resolve_depth
from .stored import StoredQuery

OWS_GET_CAPABILITIES_ELEMENTS = {
    # {"child-tag": ("item-tag", "python_name")}
    xmlns.ows11.qname("AcceptVersions"): (xmlns.ows11.qname("Version"), "acceptVersions"),
    xmlns.ows11.qname("Sections"): (xmlns.ows11.qname("Section"), "sections"),
    xmlns.ows11.qname("AcceptFormats"): (xmlns.ows11.qname("OutputFormat"), "acceptFormats"),
    xmlns.ows11.qname("AcceptLanguages"): (xmlns.ows11.qname("Language"), "acceptLanguages"),
}

# Fully qualified tag names
WFS_TYPE_NAME = xmlns.wfs20.qname("TypeName")
WFS_STORED_QUERY = xmlns.wfs20.qname("StoredQuery")


@dataclass
@tag_registry.register("GetCapabilities", xmlns.wfs20)
@tag_registry.register("GetCapabilities", xmlns.wfs1, hidden=True)  # to give negotiation error.
class GetCapabilities(BaseOwsRequest):
    """Request parsing for GetCapabilities.

    This parses and handles the syntax::

        <wfs:GetCapabilities service="WFS" updateSequence="" version="">
            <ows:AcceptVersions>
                <ows:Version>...</ows:Version>
            </ows:AcceptVersions>
            <ows:Sections>
                <ows:Section>...</ows:Section>
                <ows:Section>...</ows:Section>
            </ows:Sections>
            <ows:AcceptFormats>
                <ows:OutputFormat>...</ows:OutputFormat>
                <ows:OutputFormat>...</ows:OutputFormat>
            </ows:AcceptFormats>
            <ows:AcceptLanguages>
                <ows:Language>...</ows:Language>
                <ows:Language>...</ows:Language>
            </ows:AcceptLanguages>
        </wfs:GetCapabilities>

    And supports the GET syntax as well::

        ?SERVICE=WFS&REQUEST=GetCapabilities&ACCEPTVERSIONS=2.0.0,1.1.0
    """

    updateSequence: str | None = None
    acceptVersions: list[str] | None = None
    # Sections can be: "ServiceIdentification", "ServiceProvider", "OperationsMetadata", "FeatureTypeList", "Filter_Capabilities"
    sections: list[str] | None = None
    acceptFormats: list[str] | None = None
    acceptLanguages: list[str] | None = None

    def __post_init__(self):
        # Even GetCapabilities can still receive a version argument to fixate it.
        if self.version and self.acceptVersions:
            raise InvalidParameterValue(
                "Can't provide both AcceptVersions and version", locator="AcceptVersions"
            )

    @classmethod
    def from_xml(cls, element: NSElement):
        """Parse the XML tag for the GetCapabilities."""
        ows_kwargs = {}
        for child in element:
            pair = OWS_GET_CAPABILITIES_ELEMENTS.get(child.tag)
            if pair is not None:
                item_tag, arg_name = pair
                ows_kwargs[arg_name] = [item.text for item in child.findall(item_tag)]

        return cls(
            # version is optional for this type, unlike all other methods
            # so not calling **cls.base_xml_init_parameters()
            service=element.get_str_attribute("service"),
            version=element.attrib.get("version"),
            handle=element.attrib.get("handle"),
            updateSequence=element.attrib.get("updateSequence", None),
            **ows_kwargs,
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Parse the KVP request format."""
        return cls(
            # version is optional for this type, unlike all other methods
            # so not calling **cls.base_kvp_init_parameters()
            service=kvp.get_str("SERVICE"),
            version=kvp.get_str("VERSION", default=None),
            handle=None,
            acceptVersions=kvp.get_list("AcceptVersions", default=None),
            acceptFormats=kvp.get_list("AcceptFormats", default=None),
            acceptLanguages=kvp.get_list("AcceptLanguages", default=None),
            sections=kvp.get_list("sections", default=None),
        )


@dataclass
@tag_registry.register("DescribeFeatureType", xmlns.wfs20)
class DescribeFeatureType(BaseOwsRequest):
    """The ``<wfs:DescribeFeatureType>`` element.

    This parses the syntax::

        <wfs:DescribeFeatureType version="2.0.0" service="WFS">
          <wfs:TypeName>ns01:TreesA_1M</wfs:TypeName>
          <wfs:TypeName>ns02:RoadL_1M</wfs:TypeName>
        </wfs:DescribeFeatureType>

    And supports the GET syntax as well::

        ?SERVICE=WFS&VERSION=2.0.0&REQUEST=DescribeFeatureType&TYPENAMES=ns01:TreesA_1M,ns02:RoadL_1M
    """

    typeNames: list[str] | None = None
    # WFS spec actually defines "application/gml+xml; version="3.2" as default output format value.
    outputFormat: str | None = "XMLSCHEMA"

    @classmethod
    def from_xml(cls, element: NSElement):
        """Parse the XML POST request."""
        type_name_tags = element.findall(WFS_TYPE_NAME)
        if any(not e.text for e in type_name_tags):
            raise MissingParameterValue("Missing TypeName value", locator="TypeName")

        return cls(
            **cls.base_xml_init_parameters(element),
            typeNames=[e.parse_qname(e.text) for e in type_name_tags] or None,
            outputFormat=element.attrib.get("outputFormat", "XMLSCHEMA"),
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Parse the KVP GET request"""
        type_names = (
            # Check for empty values, don't check for missing values:
            kvp.get_list("typeNames", alias="typename")
            if ("typeNames" in kvp or "typename" in kvp)
            else None
        )

        return cls(
            **cls.base_kvp_init_parameters(kvp),
            # TYPENAME is WFS 1.x, but some clients and the Cite test suite send it.
            typeNames=(
                [kvp.parse_qname(name) for name in type_names] if type_names is not None else None
            ),
            outputFormat=kvp.get_str("outputFormat", default="XMLSCHEMA"),
        )


class ResultType(Enum):
    """WFS result type format."""

    hits = "HITS"
    results = "RESULTS"

    @classmethod
    def _missing_(cls, value):
        raise OperationParsingFailed(f"Invalid resultType value: {value}", locator="resultType")


@dataclass
class StandardPresentationParameters(BaseOwsRequest):
    """Mixin that handles presentation parameters shared between different types.
    This element mirrors the ``wfs:StandardPresentationParameters`` type.
    """

    count: int | None = None
    outputFormat: str = "application/gml+xml; version=3.2"
    resultType: ResultType = ResultType.results
    startIndex: int = 0

    @classmethod
    def base_xml_init_parameters(cls, element: NSElement) -> dict:
        """Parse the XML POST request."""
        return dict(
            **super().base_xml_init_parameters(element),
            count=element.get_int_attribute("count"),
            outputFormat=element.attrib.get("outputFormat", "application/gml+xml; version=3.2"),
            resultType=ResultType[element.attrib.get("resultType", "results")],
            startIndex=element.get_int_attribute("startIndex", 0),
        )

    @classmethod
    def base_kvp_init_parameters(cls, kvp: KVPRequest) -> dict:
        """Parse the KVP GET request."""
        return dict(
            **super().base_kvp_init_parameters(kvp),
            # maxFeatures is WFS 1.x but some clients still send it.
            count=kvp.get_int("count", alias="maxFeatures", default=None),
            outputFormat=kvp.get_str("outputFormat", default="application/gml+xml; version=3.2"),
            resultType=ResultType[kvp.get_str("resultType", default="results").lower()],
            startIndex=kvp.get_int("startIndex", default=0),
        )

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        params = super().as_kvp()
        if self.outputFormat != "application/gml+xml; version=3.2":
            params["OUTPUTFORMAT"] = self.outputFormat
        if self.resultType != ResultType.results:
            params["RESULTTYPE"] = self.resultType.value
        if self.startIndex:
            params["STARTINDEX"] = self.startIndex
        if self.count is not None:
            params["COUNT"] = self.count
        return params


@dataclass
class StandardResolveParameters(BaseOwsRequest):
    """Mixin that handles resolve parameters shared between different types.
    This element mirrors the ``wfs:StandardResolveParameters`` type.
    """

    resolve: ResolveValue = ResolveValue.none
    resolveDepth: int | None = None
    resolveTimeout: int = 300

    @classmethod
    def base_xml_init_parameters(cls, element: NSElement) -> dict:
        """Parse the XML POST request."""
        return dict(
            **super().base_xml_init_parameters(element),
            resolve=ResolveValue[element.attrib.get("resolve", "none")],
            resolveDepth=parse_resolve_depth(element.attrib.get("resolveDepth", None)),
            resolveTimeout=element.get_int_attribute("resolveTimeout", 300),
        )

    @classmethod
    def base_kvp_init_parameters(cls, kvp: KVPRequest) -> dict:
        """Parse the KVP GET request."""
        depth = kvp.get_str("resolveDepth", default="*")
        return dict(
            **super().base_kvp_init_parameters(kvp),
            resolve=ResolveValue[kvp.get_str("resolve", default="none")],
            resolveDepth=parse_resolve_depth(depth),
            resolveTimeout=kvp.get_int("resolveTimeout", default=300),
        )


@dataclass
class CommonQueryParameters(BaseOwsRequest):
    """Internal mixin to deal with the query parameters"""

    queries: list[QueryExpression] = None  # need default for inheritance

    @classmethod
    def base_xml_init_parameters(cls, element: NSElement):
        """Parse the XML POST request"""
        return dict(
            **super().base_xml_init_parameters(element),
            queries=[
                # This can instantiate an AdhocQuery or StoredQuery
                QueryExpression.child_from_xml(child)
                for child in element
            ],
        )

    @classmethod
    def base_kvp_init_parameters(cls, kvp: KVPRequest):
        """Parse the KVP GET request"""
        stored_query_id = kvp.get_str("STOREDQUERY_ID", default=None)
        if stored_query_id:
            queries = [
                StoredQuery.from_kvp_request(sub_request)
                for sub_request in kvp.split_parameter_lists()
            ]
        else:
            queries = [
                AdhocQuery.from_kvp_request(sub_request)
                for sub_request in kvp.split_parameter_lists()
            ]

        return dict(
            **super().base_kvp_init_parameters(kvp),
            queries=queries,
        )

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        if len(self.queries) > 1:
            raise NotImplementedError()
        return {
            **super().as_kvp(),
            **self.queries[0].as_kvp(),
        }


@dataclass
@tag_registry.register("GetFeature", xmlns.wfs20)
class GetFeature(
    StandardPresentationParameters,
    StandardResolveParameters,
    CommonQueryParameters,
    BaseOwsRequest,
):
    """The ``<wfs:GetFeature>`` element.

    This parses the syntax::

        <wfs:GetFeature outputFormat="application/gml+xml; version=3.2">
           <wfs:Query typeName="myns:InWaterA_1M">
              <fes:Filter>
                  ...
              </fes:Filter>
           </wfs:Query>
        </wfs:GetFeature>

    And supports the KVP syntax::

        ?SERVICE=WFS&VERSION=2.0.0&REQUEST=GetFeature&TYPENAMES=myns:InWaterA_1M&FILTER=...

    """


@dataclass
@tag_registry.register("GetPropertyValue", xmlns.wfs20)
class GetPropertyValue(
    StandardPresentationParameters,
    StandardResolveParameters,
    CommonQueryParameters,
    BaseOwsRequest,
):
    """The ``<wfs:GetPropertyValue>`` element.

    This parses the syntax::

        <wfs:GetPropertyValue valueReference="...">
           <wfs:Query typeName="myns:InWaterA_1M">
              <fes:Filter>
                  ...
              </fes:Filter>
           </wfs:Query>
        </wfs:GetFeature>

    As this is so similar to the :class:`GetFeature` syntax, these inherit from each other.

    The KVP-syntax is also supported::

        ?SERVICE=WFS&VERSION=2.0.0&REQUEST=GetPropertyName&...&&VALUEREFERENCE=...
    """

    resolvePath: str | None = None
    valueReference: fes20.ValueReference = None  # need default for inheritance

    def __post_init__(self):
        if self.resolvePath:
            raise InvalidParameterValue(
                "Support for resolvePath is not implemented!", locator="resolvePath"
            )
        if len(self.queries) > 1:
            raise InvalidParameterValue(
                "GetPropertyValue only supports a single query", locator="filter"
            )

    @classmethod
    def from_xml(cls, element: NSElement):
        """Parse the XML POST request."""
        return cls(
            **cls.base_xml_init_parameters(element),
            resolvePath=element.attrib.get("resolvePath", None),
            valueReference=fes20.ValueReference(
                xpath=element.get_str_attribute("valueReference"),
                xpath_ns_aliases=element.ns_aliases,
            ),
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Parse the KVP GET request."""
        return cls(
            **cls.base_kvp_init_parameters(kvp),
            resolvePath=kvp.get_str("resolvePath", default=None),
            valueReference=fes20.ValueReference(
                xpath=kvp.get_str("valueReference"),
                xpath_ns_aliases=kvp.ns_aliases,
            ),
        )

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        return {
            **super().as_kvp(),
            "VALUEREFERENCE": self.valueReference.xpath,
        }


@tag_registry.register("ListStoredQueries", xmlns.wfs20)
class ListStoredQueries(BaseOwsRequest):
    """The ``<wfs:ListStoredQueries>`` element."""

    @classmethod
    @expect_no_children
    def from_xml(cls, element: NSElement):
        return super().from_xml(element)


@dataclass
@tag_registry.register("DescribeStoredQueries", xmlns.wfs20)
class DescribeStoredQueries(BaseOwsRequest):
    """The ``<wfs:DescribeStoredQueries>`` element.

    This parses the syntax::

        <wfs:DescribeStoredQueries>
          <wfs:StoredQueryId>...</wfs:StoredQueryId>
          <wfs:StoredQueryId>...</wfs:StoredQueryId>
        </wfs:DescribeStoredQueries>

    And the KVP syntax::

        ?REQUEST=DescribeStoredQueries&STOREDQUERY_ID=...,...
    """

    storedQueryId: list[str] | None = None

    @classmethod
    def from_xml(cls, element: NSElement):
        """Parse the XML POST request."""
        id_tags = element.findall(WFS_STORED_QUERY)
        if any(not e.text for e in id_tags):
            raise MissingParameterValue("Missing StoredQuery value", locator="StoredQuery")

        return cls(
            **cls.base_xml_init_parameters(element),
            storedQueryId=[child.text for child in id_tags] or None,
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Parse the KVP GET request."""
        return cls(
            **cls.base_kvp_init_parameters(kvp),
            storedQueryId=kvp.get_list("STOREDQUERY_ID", default=None),
        )


# @tag_registry.register("CreateStoredQuery", xmlns.wfs20)
# class CreateStoredQuery(BaseRequest):
#     ...
#
# @tag_registry.register("LockFeature", xmlns.wfs20)
# class LockFeature(BaseRequest):
#     ...
#
# @tag_registry.register("TransactionType", xmlns.wfs20)
# class TransactionType(BaseRequest):
#     ...
#
# @tag_registry.register("DropStoredQuery", xmlns.wfs20)
# class DropStoredQuery(BaseRequest):
#     ...



"""The FES elements that handle sorting."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from gisserver.exceptions import InvalidParameterValue
from gisserver.parsers.ast import AstNode, expect_children, expect_tag, tag_registry
from gisserver.parsers.fes20 import ValueReference
from gisserver.parsers.ows import KVPRequest
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, xmlns

FES_SORT_ORDER = xmlns.fes20.qname("SortOrder")


class SortOrder(Enum):
    #: Ascending order
    ASC = ""
    #: Descrending order
    DESC = "-"

    # Support WFS 1 names for clients that still use this:

    #: WFS 1 name that clients still use for ascending.
    A = ASC
    #: WFS 1 name that clients still use for descending.
    D = DESC

    @classmethod
    def from_string(cls, direction):
        try:
            return cls[direction]
        except KeyError:
            raise InvalidParameterValue(
                "Expect ASC/DESC ordering direction", locator="sortby"
            ) from None

    @classmethod
    @expect_tag(xmlns.fes20, "SortOrder")
    def from_xml(cls, element: NSElement):
        return SortOrder.from_string(element.text)


@dataclass
@tag_registry.register("SortProperty", xmlns.fes20)
class SortProperty(AstNode):
    """This class name is based on the WFS spec.

    This parses and handles the syntax::

        <fes:SortProperty>
            <fes:ValueReference>name</fes:ValueReference>
            <fes:SortOrder>ASC</fes:SortOrder>
        </fes:SortProperty>
    """

    value_reference: ValueReference
    sort_order: SortOrder = SortOrder.ASC

    def __post_init__(self):
        if "[" in self.value_reference.xpath:
            raise InvalidParameterValue(
                "Sorting with XPath attribute selectors is not supported.", locator="sortby"
            )

    @classmethod
    @expect_tag(xmlns.fes20, "SortProperty")
    @expect_children(1, ValueReference, FES_SORT_ORDER)
    def from_xml(cls, element: NSElement) -> SortProperty:
        """Parse the incoming XML"""
        sort_order = element.find(FES_SORT_ORDER)
        return cls(
            value_reference=ValueReference.from_xml(element[0]),
            sort_order=(
                SortOrder.from_xml(sort_order) if sort_order is not None else SortOrder.ASC
            ),
        )

    @classmethod
    def from_string(cls, value: str, ns_aliases: dict[str, str]) -> SortProperty:
        """Parse the incoming GET parameter."""
        xpath, _, direction = value.partition(" ")
        return SortProperty(
            value_reference=ValueReference(xpath, ns_aliases),
            sort_order=SortOrder.from_string(direction) if direction else SortOrder.ASC,
        )

    def as_kvp(self):
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        if self.sort_order == SortOrder.ASC:
            return self.value_reference.xpath
        else:
            return f"{self.value_reference.xpath} {self.sort_order.name}"


@dataclass
@tag_registry.register("SortBy", xmlns.fes20)
class SortBy(AstNode):
    """The sortBy clause.

    This parses and handles the syntax::

        <fes:SortBy>
            <fes:SortProperty>
                <fes:ValueReference>name</fes:ValueReference>
                <fes:SortOrder>ASC</fes:SortOrder>
            </fes:SortProperty>
        </fes:SortBy>

    It also supports the SORTBY parameter for GET requests.
    """

    #: The ``<fes:SortProperty>`` elements.
    sort_properties: list[SortProperty]

    @classmethod
    @expect_children(1, SortProperty)
    def from_xml(cls, element: NSElement) -> SortBy:
        """Parse the XML tag."""
        return cls(
            sort_properties=[
                # The from_xml() validates that the child node is a <fes:SortProperty>
                SortProperty.from_xml(child)
                for child in element
            ]
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest) -> SortBy | None:
        """Construct the SortBy object from a KVP "SORTBY" parameter, and considering NAMESPACES."""
        value = kvp.get_str("SortBy", default=None)
        if value is None:
            return None

        return cls(
            sort_properties=[
                SortProperty.from_string(field, kvp.ns_aliases) for field in value.split(",")
            ]
        )

    def as_kvp(self) -> str:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        return ",".join(sort_property.as_kvp() for sort_property in self.sort_properties)

    def build_ordering(self, compiler: CompiledQuery):
        """Build the ordering for the Django ORM call."""
        ordering = []
        for prop in self.sort_properties:
            orm_path = prop.value_reference.parse_xpath(compiler.feature_types).orm_path
            ordering.append(f"{prop.sort_order.value}{orm_path}")

        compiler.add_ordering(ordering)



"""Request parsing for the common Generic Open Web Services (OWS) protocol bits."""

from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache

from django.http import QueryDict

from gisserver.exceptions import (
    ExternalParsingError,
    InvalidParameterValue,
    OperationNotSupported,
    OperationParsingFailed,
)
from gisserver.parsers.ast import AstNode, tag_registry
from gisserver.parsers.xml import NSElement, parse_xml_from_string, split_ns, xmlns

from .kvp import KVPRequest

__all__ = (
    "BaseOwsRequest",
    "resolve_kvp_parser_class",
    "resolve_xml_parser_class",
    "parse_get_request",
    "parse_post_request",
)


@dataclass
class BaseOwsRequest(AstNode):
    """Base request data for all request types of the OWS standards.
    This mirrors the ``<wfs:BaseRequestType>`` element from the WFS spec.
    """

    # dataclass limitation: when defaults are added, subclasses can't have required parameters anymore.
    service: str
    version: str | None  # none for GetCapabilities only
    handle: str | None

    @classmethod
    def from_xml(cls, element: NSElement):
        """Initialize from an XML POST request."""
        return cls(**cls.base_xml_init_parameters(element))

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Initialize from an KVP GET request."""
        return cls(**cls.base_kvp_init_parameters(kvp))

    @classmethod
    def base_xml_init_parameters(cls, element: NSElement) -> dict:
        """Parse the base attributes.
        This parses the syntax such as::

            <wfs:BaseRequest service="WFS" version="2.0.0" handle="...">

        """
        return dict(
            service=element.get_str_attribute("service"),
            version=element.get_str_attribute("version"),
            handle=element.attrib.get("handle"),
        )

    @classmethod
    def base_kvp_init_parameters(cls, kvp: KVPRequest) -> dict:
        """Parse the common Key-Value-Pair format (GET request parameters).
        This parses the syntax::

            ?SERVICE=WFS&VERSION=2.0.0
        """
        return dict(
            service=kvp.get_str("SERVICE"),
            version=kvp.get_str("VERSION"),
            handle=None,
        )

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        return {
            "SERVICE": self.service,
            "VERSION": str(self.version),
            "REQUEST": split_ns(self.xml_name)[1],
        }


@lru_cache
def _get_kvp_parsers() -> dict[str, dict[str, type[BaseOwsRequest]]]:
    """Find which classes are registered."""
    # This late initialization keeps the ability open in the future
    # to handle request types that are registered in third party code.
    ows_parsers = {}

    for xml_name, parser_class in tag_registry.find_subclasses(BaseOwsRequest).items():
        namespace, local_name = split_ns(xml_name)
        ows_parsers.setdefault(namespace, {})[local_name.upper()] = parser_class

    return ows_parsers


def resolve_kvp_parser_class(kvp: KVPRequest) -> type[BaseOwsRequest]:
    """Find the appropriate class to parse the KVP GET request data."""
    service = kvp.get_str("service").upper()
    request = kvp.get_str("request")

    # Find the appropriate request object
    service_types = _get_kvp_parsers()
    try:
        # Translating the GET parameter to a namespace makes connecting to the parser registration.
        # This doesn't take the VERSION/ACCEPTVERSIONS into account yet.
        namespace = xmlns[service.lower()].value
        request_classes = service_types[namespace]
    except KeyError:
        raise InvalidParameterValue(
            f"Unsupported service type: {service}.", locator="service"
        ) from None

    try:
        request_cls = request_classes[request.upper()]
    except KeyError:
        allowed = ", ".join(
            xml_tag for node in request_classes.values() for xml_tag in node.get_tag_names()
        )
        raise OperationNotSupported(
            f"'{request}' is not implemented, supported are: {allowed}.",
            locator="request",
        ) from None

    return request_cls


def resolve_xml_parser_class(root: NSElement) -> type[BaseOwsRequest]:
    """Find the correct class to parse the XML POST data with."""
    return tag_registry.resolve_class(root, allowed_types=(BaseOwsRequest,))


def parse_get_request(
    query_string: str | dict[str, str], ns_aliases: dict | None = None
) -> BaseOwsRequest:
    """Parse the WFS KVP GET request format into the internal request objects.
    Most code calls the resolver internally, but this variation is easier for unit testing.
    """
    if isinstance(query_string, str):
        query_string = QueryDict(query_string.lstrip("?"))

    kvp = KVPRequest(query_string, ns_aliases=ns_aliases)
    request_cls = resolve_kvp_parser_class(kvp)
    return request_cls.from_kvp_request(kvp)


def parse_post_request(xml_string: str | bytes, ns_aliases: dict | None = None) -> BaseOwsRequest:
    """Parse the XML POST request format into the internal request objects.
    Most code calls the resolver internally, but this variation is easier for unit testing.
    """
    try:
        root = parse_xml_from_string(xml_string, extra_ns_aliases=ns_aliases)
    except ExternalParsingError as e:
        raise OperationParsingFailed(f"Unable to parse XML: {e}") from e

    return tag_registry.node_from_xml(root, allowed_types=(BaseOwsRequest,))



"""Operations that implement the WFS 2.0 specification.

These operations are called by the :class:`~gisserver.views.WFSView`,
and receive the request that was parsed using :mod:`gisserver.parsers.wfs20`.
Thus, there is nearly no difference between GET/POST requests here.

In a way this looks like an MVC (Model-View-Controller) design:

* :mod:`gisserver.parsers.wfs20` parsed the request, and provides the "model".
* :mod:`gisserver.operations.wfs20` orchestrates here what to do (the controller).
* :mod:`gisserver.output` performs the rendering (the view).
"""

from __future__ import annotations

import logging
import math
import re
import typing
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from gisserver import conf, output
from gisserver.exceptions import (
    InvalidParameterValue,
    VersionNegotiationFailed,
    wrap_filter_errors,
)
from gisserver.extensions.functions import function_registry
from gisserver.extensions.queries import stored_query_registry
from gisserver.features import FeatureType
from gisserver.output import CollectionOutputRenderer
from gisserver.parsers import ows, wfs20

from .base import OutputFormat, OutputFormatMixin, Parameter, WFSOperation, XmlTemplateMixin

logger = logging.getLogger(__name__)

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")
RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.

__all__ = [
    "GetCapabilities",
    "DescribeFeatureType",
    "BaseWFSGetDataOperation",
    "GetFeature",
    "GetPropertyValue",
    "ListStoredQueries",
    "DescribeStoredQueries",
]


class GetCapabilities(XmlTemplateMixin, OutputFormatMixin, WFSOperation):
    """This operation returns map features, and available operations this WFS server supports."""

    ows_request: wfs20.GetCapabilities

    xml_template_name = "get_capabilities.xml"

    def get_output_formats(self):
        return [
            OutputFormat("text/xml", renderer_class=None),
            OutputFormat(
                # for FME (Feature Manipulation Engine)
                "application/gml+xml",
                renderer_class=None,
                version="3.2",
                title="GML",
                in_capabilities=False,
            ),
        ]

    def get_parameters(self):
        # Not calling super, as this differs slightly (not requiring VERSION)
        return [
            Parameter("service", allowed_values=list(self.view.accept_operations.keys())),
            Parameter("AcceptVersions", allowed_values=self.view.accept_versions),
            Parameter("AcceptFormats", allowed_values=[str(o) for o in self.get_output_formats()]),
        ]

    def validate_request(self, ows_request: wfs20.GetCapabilities):
        # Validate AcceptVersions
        self._parse_accept_versions(ows_request)

        # Validate AcceptFormats
        if ows_request.acceptFormats:
            for output_format in ows_request.acceptFormats:
                self.resolve_output_format(output_format, locator="AcceptFormats")

    def _parse_accept_versions(self, ows_request: wfs20.GetCapabilities) -> str | None:
        """Special parsing for the ACCEPTVERSIONS parameter."""
        if not ows_request.acceptVersions:
            return None

        matched_versions = set(ows_request.acceptVersions).intersection(self.view.accept_versions)
        if not matched_versions:
            allowed = ", ".join(self.view.accept_versions)
            raise VersionNegotiationFailed(
                f"'{','.join(ows_request.acceptVersions)}' does not contain supported versions, "
                f"supported are: {allowed}.",
                locator="acceptversions",
            )

        # Take the highest version (mapserver returns first matching)
        requested_version = sorted(matched_versions, reverse=True)[0]

        # Make sure the views+exceptions continue to operate in this version
        self.view.set_version(ows_request.service, requested_version)
        return requested_version

    def get_context_data(self) -> dict:
        # The 'service' is not read from 'params' to avoid dependency on get_parameters()
        service_operations = self.view.accept_operations[self.ows_request.service]
        app_namespaces = self.view.get_xml_namespaces_to_prefixes()

        get_feature_op_cls = service_operations["GetFeature"]
        feature_output_formats = get_feature_op_cls(self.view, None).get_output_formats()

        return {
            "xml_namespaces": app_namespaces,
            "service_description": self.view.get_service_description(self.ows_request.service),
            "accept_operations": {
                name: operation_class(self.view, None).get_parameters()
                for name, operation_class in service_operations.items()
            },
            "service_constraints": self.view.wfs_service_constraints,
            "filter_capabilities": self.view.wfs_filter_capabilities,
            "function_registry": function_registry,
            "accept_versions": self.view.accept_versions,
            "feature_types": self.view.get_bound_feature_types(),
            "feature_output_formats": feature_output_formats,
            "default_max_features": self.view.max_page_size,
            "BOUNDING_BOX": conf.GISSERVER_CAPABILITIES_BOUNDING_BOX,
        }


class DescribeFeatureType(OutputFormatMixin, WFSOperation):
    """This returns an XML Schema for the provided objects.
    Each feature is exposed as an XSD definition with its fields.
    """

    ows_request: wfs20.DescribeFeatureType

    def get_output_formats(self):
        return [
            OutputFormat("XMLSCHEMA", renderer_class=output.XmlSchemaRenderer),
            # At least one version of FME (Feature Manipulation Engine) seems to
            # send a DescribeFeatureType request with this GML as output format.
            # Do what mapserver does and just send it XML Schema.
            # This output format also seems to be used in the WFS 2.0.2 spec!
            OutputFormat(
                "application/gml+xml",
                version="3.2",
                renderer_class=output.XmlSchemaRenderer,
                in_capabilities=False,
            ),
            # OutputFormat("text/xml", subtype="gml/3.1.1"),
        ]

    def validate_request(self, ows_request: wfs20.DescribeFeatureType):
        if not ows_request.typeNames:
            self.feature_types = self.view.get_bound_feature_types()
        else:
            self.feature_types = [
                self.resolve_feature_type(type_name) for type_name in ows_request.typeNames
            ]
        self.output_format = self.resolve_output_format(ows_request.outputFormat)

    def process_request(self, ows_request: wfs20.DescribeFeatureType):
        renderer_class = self.output_format.renderer_class or output.XmlSchemaRenderer
        renderer = renderer_class(self, self.feature_types)
        return renderer.get_response()


class ListStoredQueries(WFSOperation):
    """This describes the available queries"""

    ows_request: wfs20.ListStoredQueries

    def process_request(self, ows_request: ows.BaseOwsRequest):
        renderer = output.ListStoredQueriesRenderer(
            self,
            query_descriptions=stored_query_registry.get_queries(),
        )
        return renderer.get_response()


class DescribeStoredQueries(WFSOperation):
    """This describes the available queries"""

    ows_request: wfs20.DescribeStoredQueries

    def validate_request(self, ows_request: wfs20.DescribeStoredQueries):
        if ows_request.storedQueryId is None:
            self.query_descriptions = stored_query_registry.get_queries()
        else:
            self.query_descriptions = [
                stored_query_registry.resolve_query(query_id)
                for query_id in ows_request.storedQueryId
            ]

    def process_request(self, ows_request: ows.BaseOwsRequest):
        renderer = output.DescribeStoredQueriesRenderer(self, self.query_descriptions)
        return renderer.get_response()


class BaseWFSGetDataOperation(OutputFormatMixin, WFSOperation):
    """Base class for GetFeature / GetPropertyValue"""

    ows_request: wfs20.GetFeature | wfs20.GetPropertyValue

    if typing.TYPE_CHECKING:
        # Tell the type checker subclasses will only work with CollectionOutputRenderer rendering.

        def resolve_output_format(
            self, value, locator="outputFormat"
        ) -> OutputFormat[CollectionOutputRenderer]:
            return super().resolve_output_format(value, locator=locator)

    def get_parameters(self) -> list[Parameter]:
        """Parameters to advertise in the capabilities for this method."""
        return [
            # Advertise parameters in GetCapabilities
            Parameter("resultType", allowed_values=["RESULTS", "HITS"]),
            Parameter("resolve", allowed_values=["local"]),
        ]

    def validate_request(self, ows_request: wfs20.GetFeature | wfs20.GetPropertyValue):
        """Validate the incoming data before execution."""
        self.output_format = self.resolve_output_format(ows_request.outputFormat)

        # Resolve the feature types
        # Allow these to skip, e.g. when the query needs to return 0 results (GetFeatureById).
        for query in ows_request.queries:
            type_names = query.get_type_names()
            feature_types = [
                self.resolve_feature_type(
                    type_name,
                    locator=("resourceId" if query.query_locator == "resourceId" else "typeNames"),
                )
                for type_name in type_names
            ]
            self.bind_query(query, feature_types)

            # Allow both the view and feature-type to check for access.
            # Before 2.0 only FeatureType classes offered this, which required subclassing
            # FeatureType to access view.request.user. The direct view check avoids that need.
            for feature_type in feature_types:
                self.view.check_permissions(feature_type)
                feature_type.check_permissions(self.view.request)

    def bind_query(self, query: wfs20.QueryExpression, feature_types: list[FeatureType]):
        """Allow to be overwritten in GetFeatureValue"""
        query.bind(feature_types)

    def process_request(self, ows_request: wfs20.GetFeature | wfs20.GetPropertyValue):
        """Process the query, and generate the output."""
        # Initialize the collection, which constructs the ORM querysets.
        if ows_request.resultType == wfs20.ResultType.hits:
            collection = self.get_hits()
        elif self.ows_request.resultType == wfs20.ResultType.results:
            collection = self.get_results()
        else:
            raise NotImplementedError()

        # assert False, str(collection.results[0].queryset.query)

        # Initialize the renderer.
        # This can also decorate the querysets with projection information,
        # such as converting geometries to the correct CRS or add prefetch_related logic.
        renderer = self.output_format.renderer_class(operation=self, collection=collection)

        # Fixing pagination will invoke the query,
        # hence this is done at the end
        self.set_pagination_links(collection)

        # Render it!
        return renderer.get_response()

    def get_hits(self) -> output.FeatureCollection:
        """Handle the resultType=hits query.
        This creates the QuerySet and counts the number of results.
        """
        start, count = self.get_pagination()
        results = []
        for query in self.ows_request.queries:
            with wrap_filter_errors(query):
                queryset = query.get_queryset()

            results.append(
                output.SimpleFeatureCollection(
                    source_query=query,
                    feature_types=query.feature_types,
                    queryset=queryset.none(),
                    start=start,
                    stop=start + count,  # yes, count can be passed for hits
                    number_matched=queryset.count(),
                )
            )

        return output.FeatureCollection(
            results=results,
            number_matched=sum(r.number_matched for r in results),
        )

    def get_results(self) -> output.FeatureCollection:
        """Handle the resultType=results query.
        This creates the queryset, allowing to read over all results.
        """
        start, count = self.get_pagination()
        results = []
        for query in self.ows_request.queries:
            # The querysets are not executed yet, until the output is reading them.
            with wrap_filter_errors(query):
                queryset = query.get_queryset()

            results.append(
                output.SimpleFeatureCollection(
                    source_query=query,
                    feature_types=query.feature_types,
                    queryset=queryset,
                    start=start,
                    stop=start + count,
                )
            )

        # number_matched is not given here, so some rendering formats can count it instead.
        # For GML it need to be printed at the start, but for GeoJSON it can be rendered
        # as the last bit of the response. That avoids performing an expensive COUNT query.
        return output.FeatureCollection(results=results)

    def get_pagination(self) -> tuple[int, int]:
        """Tell what the requested page size is."""
        start = max(0, self.ows_request.startIndex)

        # outputFormat.max_page_size can be math.inf to enable endless scrolling.
        # this only works when the COUNT parameter is not given.
        max_page_size = self.output_format.max_page_size or self.view.max_page_size
        default_page_size = (
            0 if self.ows_request.resultType == wfs20.ResultType.hits else max_page_size
        )
        count = min(max_page_size, self.ows_request.count or default_page_size)
        return start, count

    def set_pagination_links(self, collection: output.FeatureCollection):
        """Assign the pagination links to the collection.
        This happens within the operation logic, as it can access the original GET request.
        """
        if self.ows_request.resultType == wfs20.ResultType.hits and not self.ows_request.count:
            return

        start, count = self.get_pagination()
        stop = start + count
        if stop != math.inf:
            if start > 0:
                collection.previous = self._replace_url_params(
                    STARTINDEX=max(0, start - count),
                    COUNT=count,
                )

            # Note that reading collection.has_next will invoke the query!
            if collection.has_next:
                # TODO: fix this when returning multiple typeNames:
                collection.next = self._replace_url_params(
                    STARTINDEX=start + count,
                    COUNT=count,
                )

    def _replace_url_params(self, **updates) -> str | None:
        """Replace a query parameter in the URL"""
        new_params = self.view.request.GET.copy()  # preserve lowercase fields too
        if self.view.request.method != "GET":
            # CITE compliance testing wants to see a 'next' link for POST requests too.
            try:
                new_params.update(self.ows_request.as_kvp())
            except NotImplementedError:
                # Various POST requests can't be translated back to KVP
                # mapserver omits the 'next' link in these cases too.
                return None

        # Replace any lower/mixed case variants of the previous names:
        for name in new_params:
            upper = name.upper()
            if upper in updates:
                new_params[name] = updates.pop(upper)

        # Override/replace with new remaining uppercase variants
        new_params.update(updates)
        return f"{self.view.server_url}?{urlencode(new_params)}"


class GetFeature(BaseWFSGetDataOperation):
    """This returns all properties of the feature.

    Various query parameters allow limiting the data.
    """

    def get_output_formats(self) -> list[OutputFormat]:
        """Return the default output formats.
        This selects a different rendering depending on the ``GISSERVER_USE_DB_RENDERING`` setting.
        """
        if conf.GISSERVER_GET_FEATURE_OUTPUT_FORMATS:
            return self.get_custom_output_formats(
                conf.GISSERVER_GET_FEATURE_OUTPUT_FORMATS | conf.GISSERVER_EXTRA_OUTPUT_FORMATS
            )

        if conf.GISSERVER_USE_DB_RENDERING:
            csv_renderer = output.DBCSVRenderer
            gml32_renderer = output.DBGML32Renderer
            geojson_renderer = output.DBGeoJsonRenderer
        else:
            csv_renderer = output.CSVRenderer
            gml32_renderer = output.GML32Renderer
            geojson_renderer = output.GeoJsonRenderer

        return [
            OutputFormat(
                # Needed for cite compliance tests
                "application/gml+xml",
                version="3.2",
                renderer_class=gml32_renderer,
                title="GML",
            ),
            OutputFormat(
                "text/xml",
                subtype="gml/3.2.1",
                renderer_class=gml32_renderer,
                title="GML 3.2.1",
            ),
            # OutputFormat("gml"),
            OutputFormat(
                # identical to mapserver:
                "application/json",
                subtype="geojson",
                charset="utf-8",
                renderer_class=geojson_renderer,
                title="GeoJSON",
            ),
            OutputFormat(
                # Alias needed to make ESRI ArcGIS online accept the WFS.
                # It does not recognize the "subtype" as an alias.
                "geojson",
                renderer_class=geojson_renderer,
            ),
            OutputFormat(
                "text/csv",
                subtype="csv",
                charset="utf-8",
                renderer_class=csv_renderer,
                title="CSV",
            ),
            # OutputFormat("shapezip"),
            # OutputFormat("application/zip"),
        ] + self.get_custom_output_formats(conf.GISSERVER_EXTRA_OUTPUT_FORMATS)

    def get_custom_output_formats(self, output_formats_setting: dict) -> list[OutputFormat]:
        """Add custom output formats defined in the settings."""
        result = []
        for content_type, format_kwargs in output_formats_setting.items():
            renderer_class = format_kwargs["renderer_class"]
            if isinstance(renderer_class, str):
                format_kwargs["renderer_class"] = import_string(renderer_class)

            if not issubclass(renderer_class, CollectionOutputRenderer):
                raise ImproperlyConfigured(
                    f"The 'renderer_class' of output format {content_type!r},"
                    f" should be a subclass of CollectionOutputRenderer."
                )

            result.append(OutputFormat(content_type, **format_kwargs))

        return result


class GetPropertyValue(BaseWFSGetDataOperation):
    """This returns a limited set of properties of the feature.
    It works almost identical to GetFeature, except that it returns a single field.
    """

    ows_request: wfs20.GetPropertyValue

    def get_output_formats(self) -> list[OutputFormat]:
        """Define the output format for GetPropertyValue."""
        gml32_value_renderer = (
            output.DBGML32ValueRenderer
            if conf.GISSERVER_USE_DB_RENDERING
            else output.GML32ValueRenderer
        )

        return [
            OutputFormat(
                "application/gml+xml",
                version="3.2",
                renderer_class=gml32_value_renderer,
                title="GML",
            ),
            OutputFormat("text/xml", subtype="gml/3.2", renderer_class=gml32_value_renderer),
        ]

    def validate_request(self, ows_request: wfs20.GetPropertyValue):
        super().validate_request(ows_request)

        if ows_request.resolvePath:
            raise InvalidParameterValue(
                "Support for resolvePath is not implemented!", locator="resolvePath"
            )

    def bind_query(self, query: wfs20.QueryExpression, feature_types: list[FeatureType]):
        """Allow to be overwritten in GetFeatureValue"""
        # Pass valueReference to the query, which will include it in the FeatureProjection.
        # In the WFS-spec, the valueReference is only a presentation layer change.
        # However, in our case AdhocQuery object also performs internal processing,
        # so the query performs a "SELECT id, <fieldname>" as well.
        query.bind(feature_types, value_reference=self.ows_request.valueReference)



"""Additional ORM lookups used by the fes-filter code."""

from __future__ import annotations

from django.contrib.gis.db.models.fields import BaseSpatialField
from django.contrib.gis.db.models.lookups import DWithinLookup
from django.db import models
from django.db.models import lookups

from gisserver.compat import ArrayField


@models.CharField.register_lookup
@models.TextField.register_lookup
@models.ForeignObject.register_lookup
class FesLike(lookups.Lookup):
    """Allow ``fieldname__fes_like=...`` lookups in querysets."""

    lookup_name = "fes_like"

    def as_sql(self, compiler, connection):
        """Generate the required SQL."""
        # lhs = "table"."field"
        # rhs = %s
        # lhs_params = []
        # lhs_params = ["prep-value"]
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return f"{lhs} LIKE {rhs}", lhs_params + rhs_params

    def get_db_prep_lookup(self, value, connection):
        """This expects that the right-hand-side already has wildcard characters."""
        return "%s", [value]


@models.Field.register_lookup
@models.ForeignObject.register_lookup
class FesNotEqual(lookups.Lookup):
    """Allow ``fieldname__fes_notequal=...`` lookups in querysets."""

    lookup_name = "fes_notequal"

    def as_sql(self, compiler, connection):
        """Generate the required SQL."""
        lhs, lhs_params = self.process_lhs(compiler, connection)  # = (table.field, %s)
        rhs, rhs_params = self.process_rhs(compiler, connection)  # = ("prep-value", [])
        return f"{lhs} != {rhs}", (lhs_params + rhs_params)


@BaseSpatialField.register_lookup
class FesBeyondLookup(DWithinLookup):
    """Allow ``fieldname__fes_beyond=...`` lookups in querysets.

    Based on the FES 2.0.3 corrigendum:

    * ``DWithin(A,B,d) = Distance(A,B) < d``
    * ``Beyond(A,B,d) = Distance(A,B) > d``

    See: https://docs.opengeospatial.org/is/09-026r2/09-026r2.html#61
    """

    lookup_name = "fes_beyond"
    sql_template = "NOT %(func)s(%(lhs)s, %(rhs)s, %(value)s)"

    def get_rhs_op(self, connection, rhs):
        # Allow the SQL $(func)s to be different from the ORM lookup name.
        # This uses ST_DWithin() on PostGIS
        return connection.ops.gis_operators["dwithin"]


if ArrayField is None:
    ARRAY_LOOKUPS = {}
else:
    # Comparisons with array fields go through a separate ORM lookup expression,
    # so these can check whether ANY element matches in the array.
    # This gives consistency between other repeated elements (e.g. M2M, reverse FK)
    # where the whole object is returned when one of the sub-objects match.
    ARRAY_LOOKUPS = {
        "exact": "fes_anyexact",
        "fes_notequal": "fes_anynotequal",
        "fes_like": "fes_anylike",
        "lt": "fes_anylt",
        "lte": "fes_anylte",
        "gt": "fes_anygt",
        "gte": "fes_anygte",
    }

    class ArrayAnyMixin:
        any_operators = {
            "exact": "= ANY(%s)",
            "ne": "!= ANY(%s)",
            "gt": "< ANY(%s)",
            "gte": "<= ANY(%s)",
            "lt": "> ANY(%s)",
            "lte": ">= ANY(%s)",
        }

        def as_sql(self, compiler, connection):
            # For the ANY() comparison, the filter operands need to be reversed.
            # So instead of "field < value", it becomes "value > ANY(field)
            lhs_sql, lhs_params = self.process_lhs(compiler, connection)
            rhs_sql, rhs_params = self.process_rhs(compiler, connection)
            lhs_sql = self.get_rhs_op(connection, lhs_sql)
            return f"{rhs_sql} {lhs_sql}", (rhs_params + lhs_params)

        def get_rhs_op(self, connection, rhs):
            return self.any_operators[self.lookup_name] % rhs

    def _register_any_lookup(base: type[lookups.BuiltinLookup]):
        """Register array lookups under a different name."""
        cls = type(f"FesArrayAny{base.__name__}", (ArrayAnyMixin, base), {})
        ArrayField.register_lookup(cls, lookup_name=f"fes_any{base.lookup_name}")

    _register_any_lookup(lookups.Exact)
    _register_any_lookup(lookups.Exact)
    _register_any_lookup(lookups.GreaterThan)
    _register_any_lookup(lookups.GreaterThanOrEqual)
    _register_any_lookup(lookups.LessThan)
    _register_any_lookup(lookups.LessThanOrEqual)

    @ArrayField.register_lookup
    class FesArrayAnyNotEqual(lookups.Lookup):
        """Inequality test for a single item in the array"""

        lookup_name = "fes_anynotequal"

        def as_sql(self, compiler, connection):
            """Generate the required SQL."""
            lhs, lhs_params = self.process_lhs(compiler, connection)
            rhs, rhs_params = self.process_rhs(compiler, connection)
            return f"{rhs} != ANY({lhs})", (rhs_params + lhs_params)

    @ArrayField.register_lookup
    class FesArrayLike(FesLike):
        """Allow like lookups for array fields."""

        lookup_name = "fes_anylike"

        def as_sql(self, compiler, connection):
            """Generate the required SQL."""
            lhs, lhs_params = self.process_lhs(compiler, connection)  # = (table.field, %s)
            rhs, rhs_params = self.process_rhs(compiler, connection)  # = ("prep-value", [])
            return (
                f"EXISTS(SELECT 1 FROM unnest({lhs}) AS item WHERE item LIKE {rhs})",  # noqa: S608
                (lhs_params + rhs_params),
            )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import AnyStr, ClassVar, Union

from django.db.models import Q

from gisserver.exceptions import InvalidParameterValue
from gisserver.parsers.ast import AstNode, expect_tag, tag_registry
from gisserver.parsers.gml import GEOSGMLGeometry
from gisserver.parsers.ows import KVPRequest
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, parse_xml_from_string, xmlns

from . import expressions, identifiers, operators

#: The FES element group that can be used as body for the :class:`Filter` element.
FilterPredicates = Union[expressions.Function, operators.Operator]

# Fully qualified tag names
FES_RESOURCE_ID = xmlns.fes20.qname("ResourceId")


@dataclass
@tag_registry.register("Filter", xmlns.fes20)
class Filter(AstNode):
    """The ``<fes:Filter>`` element.

    This parses and handles the syntax::

        <fes:Filter>
            <fes:SomeOperator>
                ...
            </fes:SomeOperator>
        </fes:Filter>

    The :meth:`build_query` will convert the parsed tree
    into a format that can build a Django ORM QuerySet.

    .. seealso:: https://www.mediamaps.ch/ogc/schemas-xsdoc/sld/1.2/filter_xsd.html#Filter
    """

    query_language: ClassVar[str] = "urn:ogc:def:queryLanguage:OGC-FES:Filter"

    #: The filter predicate (body)
    predicate: FilterPredicates

    source: AnyStr | None = field(default=None, compare=False)

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest) -> Filter | None:
        """Parse the filter from the GET request."""

        # Check filter language
        filter_language = kvp.get_str("FILTER_LANGUAGE", default=cls.query_language)
        if filter_language != cls.query_language:
            raise InvalidParameterValue(
                f"Invalid value for filterLanguage: {filter_language}", locator="filterLanguage"
            )

        # Parse filter
        filter = kvp.get_custom(
            "filter", default=None, parser=lambda x: cls.from_string(x, kvp.ns_aliases)
        )

        # Parse alternatives to filter
        resource_ids = [
            identifiers.ResourceId.from_string(rid, kvp.ns_aliases)
            for rid in kvp.get_list("resourceID", default=[])
        ]
        bbox = kvp.get_custom("bbox", default=None, parser=GEOSGMLGeometry.from_bbox)

        # Make sure the various query options are not mixed.
        cls.validate_kvp_exclusions(filter, bbox, resource_ids)

        if filter is None:
            # See if the other KVP parameters still provide a basic filter.
            # Instead of implementing these parameters separately in the AdhocQueryExpression,
            # they are implemented by constructing the filter AST internally.
            if resource_ids:
                filter = Filter(predicate=operators.IdOperator(resource_ids))
            elif bbox is not None:
                filter = Filter(
                    predicate=operators.BinarySpatialOperator(
                        operatorType=operators.SpatialOperatorName.BBOX,
                        operand1=None,
                        operand2=bbox,
                    )
                )

        return filter

    @classmethod
    def validate_kvp_exclusions(
        cls, filter: Filter | None, bbox: GEOSGMLGeometry | None, resource_ids: list
    ):
        """Validate mutually exclusive parameters"""
        if filter is not None and (bbox is not None or resource_ids):
            raise InvalidParameterValue(
                "The FILTER parameter is mutually exclusive with BBOX and RESOURCEID",
                locator="filter",
            )

        if resource_ids and (bbox is not None or filter is not None):
            raise InvalidParameterValue(
                "The RESOURCEID parameter is mutually exclusive with BBOX and FILTER",
                locator="resourceId",
            )

    @classmethod
    def from_string(cls, text: AnyStr, ns_aliases: dict[str, str] | None = None) -> Filter:
        """Parse an XML ``<fes:Filter>`` string.

        This uses defusedxml by default, to avoid various XML injection attacks.

        :raises ValueError: When data is incorrect, or XML has syntax errors.
        :raises NotImplementedError: When unsupported features are called.
        """
        if isinstance(text, str):
            end_first = text.index(">")
            first_tag = text[:end_first].lstrip()
            # Allow KVP requests without a namespace
            # Both geoserver and mapserver support this.
            if "xmlns" not in first_tag and (
                first_tag == "<Filter" or first_tag.startswith("<Filter ")
            ):
                text = f'{first_tag} xmlns="{xmlns.fes20}" xmlns:gml="{xmlns.gml32}"{text[end_first:]}'

        root_element = parse_xml_from_string(text, extra_ns_aliases=ns_aliases)
        return Filter.from_xml(root_element, source=text)

    @classmethod
    @expect_tag(xmlns.fes20, "Filter")
    def from_xml(cls, element: NSElement, source: AnyStr | None = None) -> Filter:
        """Parse the <fes20:Filter> element."""
        if len(element) > 1 or element[0].tag == FES_RESOURCE_ID:
            # fes20:ResourceId is the only element that may appear multiple times.
            # Wrap it in an IdOperator so this class can have a single element as predicate.
            return Filter(
                predicate=operators.IdOperator(
                    [identifiers.Id.child_from_xml(child) for child in element]
                ),
                source=source,
            )
        else:
            return Filter(
                # Can be Function or Operator (e.g. BinaryComparisonOperator),
                # but not Literal or ValueReference.
                predicate=tag_registry.node_from_xml(
                    element[0], allowed_types=FilterPredicates.__args__
                ),
                source=source,
            )

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        """Collect the data to perform a Django ORM query."""
        # Function, Operator, IdList
        # The operators may add the logic themselves, or return a Q object.
        return self.predicate.build_query(compiler)

    def get_resource_id_types(self) -> list[str] | None:
        """When the filter predicate consists of ``<fes:ResourceId>`` elements, return those.
        This can return an empty list in case a ``<fes:ResourceId>`` object doesn't define a type.
        """
        if isinstance(self.predicate, operators.IdOperator):
            return self.predicate.get_type_names()
        else:
            return None



"""Storage and registry for stored queries.
These definitions follow the WFS spec.

By using the :attr:`stored_query_registry`, custom stored queries can be registered in this server.
Out of the box, only the mandatory built-in :class:`GetFeatureById` query is present.
"""

from __future__ import annotations

import typing
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field
from functools import partial
from xml.etree.ElementTree import Element

from django.db.models import Q

from gisserver.exceptions import InvalidParameterValue, NotFound, OperationNotSupported
from gisserver.features import FeatureType
from gisserver.parsers import fes20
from gisserver.parsers.query import CompiledQuery
from gisserver.types import XsdTypes

if typing.TYPE_CHECKING:
    from gisserver.output import SimpleFeatureCollection

__all__ = (
    "GetFeatureById",
    "QueryExpressionText",
    "StoredQueryDescription",
    "StoredQueryRegistry",
    "stored_query_registry",
    "WFS_LANGUAGE",
    "FES_LANGUAGE",
)

#: The query body is a ``<wfs:Query>``.
WFS_LANGUAGE = "urn:ogc:def:queryLanguage:OGC-WFS::WFS_QueryExpression"
#: The query body is a ``<fes:Filter>``.
FES_LANGUAGE = "urn:ogc:def:queryLanguage:OGC-FES:Filter"


@dataclass
class QueryExpressionText:
    """Define the body of a stored query.

    This object type is defined in the WFS spec.
    It may contain a ``<wfs:Query>`` or ``<fes:Filter>`` element.
    """

    #: Which types the query will return.
    return_feature_types: list[str] | None = None
    #: The internal language of the query. Can be a WFS/FES-filter, or "python"
    language: str = FES_LANGUAGE
    #: Whether the implementation_text will be show or hidden from users.
    is_private: bool = True

    #: Body
    implementation_text: str | Element | None = None


@dataclass
class StoredQueryDescription:
    """WFS metadata of a stored query.
    This is based on the ``<wfs:StoredQueryDescription>`` element,
    and returned in ``DescribeStoredQueries``.

    While it's possible to define multiple :class:`QueryExpressionText` nodes
    as metadata to describe a query, there is still only one implementation.
    Note there is no 'typeNames=...' parameter for stored queries.
    Only direct parameters act as input.
    """

    #: The ID of the query
    id: str
    #: User-visible title
    title: str
    #: User-visible description
    abstract: str
    #: Parameter declaration
    parameters: dict[str, XsdTypes]

    #: Metadata describing the query body
    expressions: list[QueryExpressionText] = field(
        default_factory=lambda: [QueryExpressionText(language=FES_LANGUAGE)]
    )

    #: Python-based implementation for the query.
    implementation_class: type[StoredQueryImplementation] = field(init=False, default=None)


class StoredQueryImplementation:
    """A custom stored query.

    This receives the parameters as init arguments,
    and should implement :meth:`build_query`.
    The function is registered using ``StoredQueryRegistry.register()``.
    """

    # Registered metadata
    _meta: StoredQueryDescription

    # Allow queries to return only the XML nodes, without any <wfs:FeatureCollection> wrapper.
    has_standalone_output: bool = False

    def __repr__(self):
        return f"<{self.__class__.__name__} implementing '{self._meta.id}'>"

    def bind(self, source_query, feature_types: list[FeatureType]):
        """Associate this query with the application data."""
        self.source_query = source_query
        self.feature_types = feature_types
        if len(feature_types) > 1:
            raise OperationNotSupported("Join queries are not supported", locator="typeNames")

    def get_type_names(self) -> list[str]:
        """Tell which type names this query applies to."""
        raise NotImplementedError()

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        """Contribute our filter expression to the internal query.

        This should add the filter expressions to the internal query compiler.
        The top-level ``<wfs:StoredQuery>`` object will add
        the ``<wfs:PropertyName>`` logic and other elements.
        """
        raise NotImplementedError()

    def finalize_results(self, result: SimpleFeatureCollection):
        """Hook to allow subclasses to inspect the results."""


class StoredQueryRegistry:
    """Registry of functions to be callable by ``<wfs:StoredQuery>`` and ``STOREDQUERY_ID=...``."""

    def __init__(self):
        self.stored_queries: dict[str, type(StoredQueryImplementation)] = {}

    def __bool__(self):
        return bool(self.stored_queries)

    def __iter__(self) -> Iterator[StoredQueryDescription]:
        return iter(self.stored_queries.values())

    def get_queries(self) -> Iterable[StoredQueryDescription]:
        """Find all descriptions for stored queries."""
        return self.stored_queries.values()

    def register(
        self,
        meta: StoredQueryDescription | None = None,
        query_expression: type[StoredQueryImplementation] | None = None,
        **meta_kwargs,
    ):
        """Register a custom class that handles a stored query.
        This function can be used as decorator or normal call.
        """
        if meta is None:
            meta = StoredQueryDescription(**meta_kwargs)
        elif meta_kwargs:
            raise TypeError("Either provide the 'meta' object or 'meta_kwargs'")

        if query_expression is not None:
            return self._register(meta, query_expression)
        else:
            return partial(self._register, meta)  # decorator effect.

    def _register(
        self, meta: StoredQueryDescription, implementation_class: type[StoredQueryImplementation]
    ):
        """Internal registration method."""
        if not issubclass(implementation_class, StoredQueryImplementation):
            raise TypeError(f"Expecting {StoredQueryImplementation}' subclass")
        if meta.implementation_class is not None:
            raise RuntimeError("Can't register same StoredQueryDescription again.")

        # for now link both. There is always a single implementation for the metadata.
        meta.implementation_class = implementation_class
        self.stored_queries[meta.id] = meta
        return implementation_class  # for decorator usage

    def resolve_query(self, query_id) -> type[StoredQueryDescription]:
        """Find the stored procedure using the ID."""
        try:
            return self.stored_queries[query_id]
        except KeyError:
            raise InvalidParameterValue(
                f"Stored query does not exist: {query_id}",
                locator="STOREDQUERY_ID",
            ) from None


#: The stored query registry
stored_query_registry = StoredQueryRegistry()


@stored_query_registry.register(
    id="urn:ogc:def:query:OGC-WFS::GetFeatureById",
    title="Get feature by identifier",
    abstract="Returns the single feature that corresponds with the ID argument",
    parameters={"id": XsdTypes.string},
    expressions=[QueryExpressionText(language=WFS_LANGUAGE)],
)
class GetFeatureById(StoredQueryImplementation):
    """The stored query for GetFeatureById.

    This can be called using::

        <wfs:StoredQuery id="urn:ogc:def:query:OGC-WFS::GetFeatureById">
            <wfs:Parameter name="ID">typename.ID</wfs:Parameter>
        </wfs:StoredQuery>

    or using KVP syntax::

        ?...&REQUEST=GetFeature&STOREDQUERY_ID=urn:ogc:def:query:OGC-WFS::GetFeatureById&ID=typename.ID

    The execution of the ``GetFeatureById`` query is essentially the same as::

        <wfs:Query xmlns:wfs="..." xmlns:fes="...'>
          <fes:Filter><fes:ResourceId rid='{ID}'/></fes:Filter>
        </wfs:Query>

    or::

        <wfs:Query typeName="{typename}">
          <fes:Filter>
            <fes:PropertyIsEqualTo>
              <fes:ValueReference>{primary-key-field}</fes:ValueReference>
              <fes:Literal>{ID-value}</fes:Literal>
            </fes:PropertyIsEqualTo>
          </fes:Filter>
        </wfs:Query>

    Except that the response is supposed to contain only the item itself.
    """

    # Projection of GetFeatureById only returns the document nodes, not a <wfs:FeatureCollection> wrapper
    has_standalone_output = True

    def __init__(self, id: str, ns_aliases: dict[str, str]):
        """Initialize the query with the request parameters."""
        if "." not in id:
            # Always report this as 404
            raise NotFound("Expected typeName.id for ID parameter", locator="ID") from None

        # GetFeatureById is essentially a ResourceId lookup, reuse that logic here.
        self.resource_id = fes20.ResourceId.from_string(id, ns_aliases)

    def get_type_names(self) -> list[str]:
        """Tell which type names this query applies to."""
        return [self.resource_id.get_type_name()]

    def build_query(self, compiler: CompiledQuery) -> Q:
        """Contribute our filter expression to the internal query."""
        try:
            return self.resource_id.build_query(compiler)
        except InvalidParameterValue as e:
            raise InvalidParameterValue(f"Invalid ID value: {e.__cause__}", locator="ID") from e

    def finalize_results(self, results: SimpleFeatureCollection):
        """Override to implement 404 checking."""
        # Directly attempt to collect the data.
        # Avoid having to do that in the output renderer.
        if results.first() is None:
            # WFS 2.0.2: Return NotFound instead of InvalidParameterValue
            raise NotFound(f"Feature not found with ID {self.resource_id.rid}.", locator="ID")



"""Parsing of scalar values in the request."""

import logging
import re
from datetime import date, datetime, time
from decimal import Decimal as D

from django.utils.dateparse import parse_date, parse_datetime, parse_duration, parse_time

from gisserver.exceptions import ExternalParsingError

logger = logging.getLogger(__name__)
RE_FLOAT = re.compile(r"\A[0-9]+(\.[0-9]+)\Z")


def auto_cast(value: str):
    """Automatically cast a value to a scalar.

    This recognizes integers, floats and ISO datetimes.
    Booleans are not handled, as that leads to unpredictable behavior
    (as seen in Yaml for example).
    """
    if not isinstance(value, str):
        return value

    if value.isdigit():
        return int(value)
    elif RE_FLOAT.match(value):
        return D(value)
    elif "T" in value:
        try:
            return parse_iso_datetime(value)
        except ValueError:
            pass

    return value


def parse_iso_date(raw_value: str) -> date:
    """Translate ISO date into a Python date value."""
    try:
        value = parse_date(raw_value)
    except ValueError as e:
        raise ExternalParsingError(str(e)) from e

    if value is None:
        raise ExternalParsingError("Date must be in YYYY-MM-DD format.")
    return value


def parse_iso_datetime(raw_value: str) -> datetime:
    """Translate ISO datetimes into a Python datetime value."""
    try:
        value = parse_datetime(raw_value)
    except ValueError as e:
        raise ExternalParsingError(str(e)) from e

    if value is None:
        raise ExternalParsingError("Date must be in YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ] format.")
    return value


def parse_iso_time(raw_value: str) -> time:
    """Translate ISO times into a Python time value."""
    try:
        value = parse_time(raw_value)
    except ValueError as e:
        raise ExternalParsingError(str(e)) from e

    if value is None:
        raise ExternalParsingError("Time must be in HH:MM[:ss[.uuuuuu]][TZ] format.")
    return value


def parse_iso_duration(raw_value: str) -> time:
    """Translate ISO times into a Python time value."""
    # The parse_duration() supports multiple formats, including ISO8601.
    # Limit it to only one format.
    if not raw_value.startswith(("P", "-P")):
        raise ExternalParsingError("Duration must be in ISO8601 format (e.g. PT1H).")

    try:
        value = parse_duration(raw_value)
    except ValueError as e:
        raise ExternalParsingError(str(e)) from e

    if value is None:
        raise ExternalParsingError("Duration must be in ISO8601 format (e.g. PT1H).")
    return value


def parse_bool(raw_value: str):
    """Translate XML notations of true/1 and false/0 into a boolean."""
    if raw_value in ("true", "1"):
        return True
    elif raw_value in ("false", "0"):
        return False
    else:
        raise ExternalParsingError(f"Can't cast '{raw_value}' to boolean")


def fix_type_name(type_name: str, feature_namespace: str):
    """Fix the XML namespace for a typename value.

    When the default namespace points to the "wfs" or "gml" namespaces,
    parsing the QName of the type value will resolve that element as existing there.
    This will correct such error, and restore the feature-type namespace.
    """
    if feature_namespace[0] == "{":
        raise ValueError("Incorrect namespace argument")

    alt_type_name = type_name
    if type_name.startswith("{http://www.opengis.net/"):
        # When the XML POST request used xmlns="http://www.opengis.net/wfs/2.0",
        # this will define a default namespace that all QName values resolve to.
        # As this is not detectable at the moment of parsing, correct it here.
        alt_type_name = f"{{{feature_namespace}}}{type_name[type_name.index('}')+1:]}"
        logger.debug("Corrected namespaced '%s' to '%s'", type_name, alt_type_name)
    elif type_name[0] != "{":
        # Typically happens in GET requests, or when no namespace prefix is used in XML POST
        alt_type_name = f"{{{feature_namespace}}}{type_name}"
        logger.debug("Corrected unnamespaced '%s' to namespaced '%s'", type_name, alt_type_name)
    return alt_type_name



"""The base protocol to implement an operation.

The request itself is parsed by :mod:`gisserver.parsers.wfs20`, and handled here.
It can be seen as the "controller" that handles the actual request type.

All operations extend from an WFSOperation class.
This defines the metadata for the ``GetCapabilities`` call and possible output formats.

Each :class:`WFSOperation` can define a :attr:`~WFSOperation.parser_class`, or let it autodetect.
"""

from __future__ import annotations

import logging
import math
import re
import typing
from dataclasses import dataclass
from functools import cached_property

from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponse
from django.template.loader import render_to_string

from gisserver.exceptions import InvalidParameterValue
from gisserver.features import FeatureType
from gisserver.output.base import OutputRenderer
from gisserver.parsers import ows
from gisserver.parsers.values import fix_type_name

if typing.TYPE_CHECKING:
    from gisserver.views import WFSView

logger = logging.getLogger(__name__)
NoneType = type(None)
R = typing.TypeVar("R", bound=OutputRenderer)

RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.

__all__ = (
    "Parameter",
    "OutputFormat",
    "WFSOperation",
    "OutputFormatMixin",
    "XmlTemplateMixin",
)


@dataclass(frozen=True)
class Parameter:
    """The ``<ows:Parameter>`` tag to output in ``GetCapabilities``."""

    #: The name of the parameter
    name: str
    #: The values to report in ``<ows:AllowedValues><ows:Value>``.
    allowed_values: list[str]


class OutputFormat(typing.Generic[R]):
    """Declare an output format for the method.

    These formats are used in the :meth:`get_output_formats` for
    any :class:`WFSOperation` that implements :class:`OutputFormatMixin`.
    This also connects the output format with a :attr:`renderer_class`.
    """

    #: The class that performs the output rendering.
    renderer_class: type[R] | None = None

    def __init__(
        self,
        content_type,
        *,
        subtype=None,
        renderer_class: type[R] | None = None,  # None is special case for GetCapabilities
        max_page_size=None,
        title=None,
        in_capabilities=True,
        **extra,
    ):
        """
        :param content_type: The MIME-type used in the request to select this type.
        :param subtype: Shorter alias for the MIME-type.
        :param renderer_class: The class that performs the output rendering.
        :param max_page_size: Used to override the ``max_page_size`` of the renderer_class.
        :param title: A human-friendly name for an HTML overview page.
        :param in_capabilities: Whether this format needs to be advertised in GetCapabilities.
        :param extra: Any additional key-value pairs for the definition.
        """
        self.content_type = content_type
        self.subtype = subtype
        self.extra = extra
        self.renderer_class = renderer_class
        self.title = title
        self.in_capabilities = in_capabilities
        self._max_page_size = max_page_size

    def matches(self, value):
        """Test whether the 'value' is matched by this object."""
        if self.content_type == "application/gml+xml":
            # Allow "application/gml+xml; version=3.2" as sent by FME.
            # TODO rewrite this matching code in a clean way.
            value = value.split("; ", 1)[0]
        return self.content_type == value or self.subtype == value

    @property
    def identifier(self):
        """The identifier to use in templates as OUTPUTFORMAT input value."""
        return self.subtype or self.content_type

    @property
    def max_page_size(self):
        """Override the default max page size"""
        if self._max_page_size:
            return self._max_page_size
        elif self.renderer_class is not None:
            return self.renderer_class.max_page_size
        else:
            return None

    @property
    def has_infinite_page_size(self):
        """Return whether the output format can be unpaginated."""
        return self.max_page_size == math.inf

    def __str__(self):
        extra = "".join(f"; {name}={value}" for name, value in self.extra.items())
        if self.subtype:
            extra = f"; subtype={self.subtype}{extra}"
        return f"{self.content_type}{extra}"

    def __repr__(self):
        return f"<OutputFormat: {self}>"


class WFSOperation:
    """Basic interface to implement an WFS method.

    Each operation in this GIS-server extends from this base class. This class
    also exposes all requires metadata for the GetCapabilities request.
    """

    #: Optionally, mention explicitly what the parser class should be used.
    #: Otherwise, it's automatically resolved from the registered types.
    parser_class: type[ows.BaseOwsRequest] = None

    def __init__(self, view: WFSView, ows_request: ows.BaseOwsRequest):
        self.view = view
        self.ows_request = ows_request

    def get_parameters(self) -> list[Parameter]:
        """Parameters to advertise in the capabilities for this method."""
        return [
            # Always advertise SERVICE and VERSION as required parameters:
            Parameter("service", allowed_values=list(self.view.accept_operations.keys())),
            Parameter("version", allowed_values=self.view.accept_versions),
        ]

    def validate_request(self, ows_request: ows.BaseOwsRequest):
        """Validate the request."""

    def process_request(self, ows_request: ows.BaseOwsRequest):
        """Default call implementation: render an XML template."""
        raise NotImplementedError()

    @cached_property
    def all_feature_types_by_name(self) -> dict[str, FeatureType]:
        """Create a lookup for feature types by name.
        This can be cached as the :class:`WFSOperation` is instantiated with each request.
        """
        return {ft.xml_name: ft for ft in self.view.get_bound_feature_types()}

    def resolve_feature_type(self, type_name: str, locator: str = "typeNames") -> FeatureType:
        """Find the FeatureType defined in the application that corresponds with the XML type name."""
        alt_type_name = fix_type_name(type_name, self.view.xml_namespace)
        try:
            return self.all_feature_types_by_name[alt_type_name]
        except KeyError:
            if alt_type_name:
                logger.debug(
                    "Unable to locate '%s' (nor %s) in the server, options are: %r",
                    type_name,
                    alt_type_name,
                    list(self.all_feature_types_by_name),
                )
            else:
                logger.debug(
                    "Unable to locate '%s' in the server, options are: %r",
                    type_name,
                    list(self.all_feature_types_by_name),
                )
            raise InvalidParameterValue(
                f"Typename '{type_name}' doesn't exist in this server.",
                locator=locator or "typeNames",
            ) from None


class XmlTemplateMixin:
    """Mixin to support methods that render using a template."""

    #: Default template to use for rendering
    #: This is resolved as :samp:`gisserver/{service}/{version}/{xml_template_name}`.
    xml_template_name = None

    #: The content-type to render.
    xml_content_type = "text/xml; charset=utf-8"

    def process_request(self, ows_request: ows.BaseOwsRequest):
        """Process the request by rendering a Django template."""
        context = self.get_context_data()
        return self.render_xml(context, ows_request)

    def get_context_data(self):
        """Collect all arguments to use for rendering the XML template"""
        return {}

    def render_xml(self, context, ows_request: ows.BaseOwsRequest):
        """Render the response using a template."""
        return HttpResponse(
            render_to_string(
                self._get_xml_template_name(ows_request),
                context={
                    "view": self.view,
                    "app_xml_namespace": self.view.xml_namespace,
                    "server_url": self.view.server_url,
                    **context,
                },
            ),
            content_type=self.xml_content_type,
        )

    def _get_xml_template_name(self, ows_request: ows.BaseOwsRequest) -> str:
        """Generate the XML template name for this operation, and check its file pattern"""
        service = ows_request.service.lower()
        template_name = f"gisserver/{service}/{self.view.version}/{self.xml_template_name}"

        # Since 'service' and 'version' are based on external input,
        # these values are double-checked again to avoid remove file inclusion.
        if not RE_SAFE_FILENAME.match(service) or not RE_SAFE_FILENAME.match(self.view.version):
            raise SuspiciousOperation(f"Refusing to render template name {template_name}")

        return template_name


class OutputFormatMixin:
    """Mixin to support methods that handle different output formats."""

    def get_output_formats(self) -> list[OutputFormat]:
        """List all output formats. This is exposed in GetCapabilities,
        and used for internal rendering.
        """
        raise NotImplementedError()

    def resolve_output_format(self, value, locator="outputFormat") -> OutputFormat:
        """Select the proper OutputFormat object based on the input value"""
        # When using ?OUTPUTFORMAT=application/gml+xml", it is actually a URL-encoded space
        # character. Hence, spaces are replaced back to a '+' character to allow such notation
        # instead of forcing it to be ?OUTPUTFORMAT=application/gml%2bxml".
        for v in {value, value.replace(" ", "+")}:
            for o in self.get_output_formats():
                if o.matches(v):
                    return o
        raise InvalidParameterValue(
            f"'{value}' is not a permitted output format for this operation.",
            locator=locator or "outputFormat",
        ) from None



"""Quick utility to import GeoJSON data in the Django project."""

from __future__ import annotations

import json
import operator
from argparse import ArgumentTypeError
from functools import reduce
from itertools import islice
from urllib.request import urlopen

from django.apps import apps
from django.contrib.gis.db.models import GeometryField
from django.contrib.gis.geos import GEOSGeometry
from django.core.exceptions import FieldDoesNotExist
from django.core.management import BaseCommand, CommandError, CommandParser
from django.db import DEFAULT_DB_ALIAS, connections, models, transaction

from gisserver.crs import CRS, CRS84


def _parse_model(value):
    try:
        return apps.get_model(value)
    except LookupError as e:
        raise ArgumentTypeError(str(e)) from e


def _parse_fields(value):
    field_map = {}
    for pair in value.split(","):
        geojson_name, _, field_name = pair.strip().partition("=")
        if not field_name:
            raise ArgumentTypeError("Expect property=field,property2=field2 format")
        field_map[geojson_name] = field_name
    return field_map


class Command(BaseCommand):
    """Quick command to import data in the WFS server."""

    help = (
        "Import GeoJSON data into a WFS feature collection. This can be done using:"
        "  manage.py loadgeojson --model=places.Province"
        " -f name=name,prop2=field2 --geometry-field geometry provinces.json"
    )

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            choices=tuple(connections),
            help=(
                "Nominates a specific database to load fixtures into. Defaults to the "
                '"default" database.'
            ),
        )
        parser.add_argument(
            "-m",
            "--model",
            required=True,
            metavar="MODEL",
            type=_parse_model,
            help="Django model, in the format: app_label.ModelName.",
        )
        parser.add_argument(
            "-g",
            "--geometry-field",
            required=False,
            metavar="NAME",
            help="Name of the model's geometry field, auto-detected if not specified.",
        )
        parser.add_argument(
            "-f",
            "--field",
            action="append",
            dest="map_fields",
            type=_parse_fields,
            metavar="NAME=FIELD",
            help=(
                "Map GeoJSON properties to Django fields, in the format: property1=field1,property2=field2,... "
                "By default, this autodetects common field names only. "
                "Explicitly ignore fields by using -f property=,property2=field2."
            ),
        )
        parser.add_argument(
            "geojson-file",
            help="GeoJSON file or URL to import. For debugging, it's better to download the file first.",
        )

    def handle(self, *args, **options):
        # Get arguments
        self.using = options["database"]
        self.connection = connections[self.using]
        model: type[models.Model] = options["model"]
        main_geometry_field = self._get_geometry_field(model, options["geometry_field"])
        field_map = self._parse_field_map(model, options["map_fields"])

        # Read the file
        geojson = self._load_geojson(options["geojson-file"])
        if not geojson["features"]:
            self.stdout.write(self.style.NOTICE("Empty GeoJSON data"))
            return

        # See if properties match the Django field names, use those too.
        # (unless these are mapped already via the command line args).
        field_map.update(self._get_auto_field_map(model, geojson["features"][0], field_map))

        # See if a CRS is declared
        self.crs = self._read_crs(geojson)

        # Import in chunks
        num_imported = 0
        id_field = model._meta.pk.name
        with transaction.atomic(using=self.using):
            features = iter(self._read_geojson(geojson, model, main_geometry_field, field_map))
            while batch := list(islice(features, 100)):
                if id_field in batch[0]:
                    # The ID field is provided, allow "on conflict update..."
                    unique_fields = [id_field]
                    update_fields = list(batch[0].keys())
                    update_fields.remove(id_field)

                    model.objects.using(self.using).bulk_create(
                        [model(**values) for values in batch],
                        update_conflicts=True,
                        unique_fields=unique_fields,
                        update_fields=update_fields,
                    )
                else:
                    model.objects.using(self.using).bulk_create(
                        [model(**values) for values in batch]
                    )

                num_imported += len(batch)

        self.stdout.write(f"Installed {num_imported} feature(s)")

    def _get_geometry_field(self, model: type[models.Model], field_name: str | None):
        """Find the geometry field name, or validate it."""
        if field_name:
            # Field is given in CLI args, validate it.
            try:
                field = model._meta.get_field(field_name)
            except FieldDoesNotExist as e:
                raise CommandError(f"Invalid value for --geometry-field: {e}") from e

            if not isinstance(field, GeometryField):
                raise CommandError("Invalid value for --geometry-field: not a GeometryField")
        else:
            # Field is not given, auto-detect
            try:
                field_name = next(
                    f.name for f in model._meta.fields if isinstance(f, GeometryField)
                )
            except StopIteration:
                raise CommandError(
                    f"Model {model._meta.label} does not have any GeometryField member."
                ) from None

        return field_name

    def _load_geojson(self, filename) -> dict:
        """Parse and validate the GeoJSON data into Python dict data."""
        try:
            if "://" in filename:
                with urlopen(filename, timeout=60) as response:  # noqa: S310
                    geojson = json.load(response)
            else:
                with open(filename) as fh:
                    geojson = json.load(fh)
        except OSError as e:  # FileNotFoundError or HTTP errors
            raise CommandError(str(e)) from e
        except (ValueError, TypeError) as e:
            raise CommandError(f"Unable to parse GeoJSON: {e}") from e

        if (
            not isinstance(geojson, dict)
            or geojson.get("type") != "FeatureCollection"
            or "features" not in geojson
        ):
            raise CommandError("Invalid GeoJSON data, expected FeatureCollection element.")

        return geojson

    def _read_crs(self, geojson: dict) -> CRS:
        """Find the CRS that should be used for all geometry data."""
        crs = geojson.get("crs")
        if not crs:
            return CRS84  # default for GeoJSON

        try:
            type = crs["type"]
            if type == "name":
                return CRS.from_string(crs["properties"]["name"])
            else:
                # type 'link' with subtype 'proj4/ogcwkt/esriwkt' are not handled here.
                raise CommandError(f"CRS type {type} is not supported.")
        except KeyError as e:
            raise CommandError(f"CRS is invalid, missing '{e}' field") from e

    def _parse_field_map(self, model, map_fields: list[dict]) -> dict:
        """Validate that the provided field mapping points to model fields."""
        if not map_fields:
            return {}

        field_map: dict = reduce(operator.or_, map_fields)
        for field_name in field_map.values():
            try:
                model._meta.get_field(field_name)
            except FieldDoesNotExist as e:
                raise CommandError(f"Field '{field_name}' does not exist in Django model.") from e

        return field_map

    def _get_auto_field_map(self, model, feature: dict, field_map: dict):
        """Autodetect which properties match model field names."""
        auto_field_map = {}
        for geojson_name in feature.get("properties", ()):
            if geojson_name in field_map:
                continue

            try:
                model._meta.get_field(geojson_name)
            except FieldDoesNotExist:
                msg = (
                    f"GeoJSON property '{geojson_name}' is not mapped nor exists in Django model, skipping."
                    if field_map
                    else f"GeoJSON property '{geojson_name}' does not exist in Django model, skipping."
                )
                self.stderr.write(self.style.WARNING(msg))
            else:
                self.stdout.write(
                    f"GeoJSON property '{geojson_name}' also exists as Django model field, using."
                )
                auto_field_map[geojson_name] = geojson_name

        return auto_field_map

    def _read_geojson(
        self, geojson: dict, model: type[models.Model], main_geometry_field: str, field_map: dict
    ):
        """Convert the GeoJSON data to model field names."""
        pk_field = model._meta.pk.name

        for feature in geojson["features"]:
            # Validate basic layout
            try:
                feature_type = feature["type"]
                geometry_data = feature["geometry"]
            except KeyError as e:
                raise CommandError(f"Feature does not have required '{e}' element.") from e
            if feature_type != "Feature":
                raise CommandError(f"Expected 'Feature', not {feature_type}")

            # Collect all fields
            properties = feature.get("properties", {})
            field_values = {
                field_name: self._parse_value(
                    model._meta.get_field(field_name), properties.get(geojson_name)
                )
                for geojson_name, field_name in field_map.items()
                if field_name
            }

            # Store the geometry, note GEOS only parses stringified JSON.
            if geometry_data is None:
                geometry = None
            else:
                try:
                    # NOTE: this looses the axis ordering information,
                    # and assumes x/y as the standard prescribes.
                    geometry = GEOSGeometry(json.dumps(geometry_data), srid=self.crs.srid)
                except ValueError as e:
                    raise CommandError(
                        f"Unable to parse geometry data: {geometry_data!r}: {e}"
                    ) from e
                geometry.srid = self.crs.srid  # override default

            field_values[main_geometry_field] = geometry

            # Try to decode the identifier if it's present
            if (
                pk_field not in field_values
                and (raw_value := feature.get("id")) is not None
                and (id_value := self._parse_id(model._meta.pk, raw_value)) is not None
            ):
                field_values[pk_field] = id_value

            # Pass as Django model fields
            yield field_values

    def _parse_id(self, pk_field: models.Field, id_value):
        # Allow TypeName.id format, see if it parses
        try:
            return pk_field.get_db_prep_save(
                id_value.rpartition(".")[2], connection=self.connection
            )
        except (ValueError, TypeError):
            self.stderr.write(
                self.style.WARNING(
                    f"Feature id value '{id_value}' can't be stored in the primary key field. ignoring."
                )
            )
            return None

    def _parse_value(self, field: models.Field, value):
        try:
            return field.get_db_prep_save(value, connection=self.connection)
        except (ValueError, TypeError) as e:
            raise CommandError(f"Can't parse {value!r} in model field '{field.name}': {e}.") from e



"""These classes map to the FES 2.0 specification for identifiers.
The class names are identical to those in the FES spec.

Inheritance structure:

* :class:`Id`

 * :class:`ResourceId`
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from django.db.models import Q

from gisserver import conf
from gisserver.exceptions import ExternalValueError, InvalidParameterValue
from gisserver.parsers.ast import AstNode, expect_no_children, expect_tag, tag_registry
from gisserver.parsers.values import auto_cast, parse_iso_datetime
from gisserver.parsers.xml import parse_qname, xmlns

NoneType = type(None)


class VersionActionTokens(Enum):
    """Values for the 'version' attribute of the :class:`ResourceId` node."""

    FIRST = "FIRST"
    LAST = "LAST"
    ALL = "ALL"
    NEXT = "NEXT"
    PREVIOUS = "PREVIOUS"


class Id(AstNode):
    """Abstract base class, as defined by FES spec.
    Any custom identifier-element needs to extend from this node.
    By default, the :class:`ResourceId` element is supported.
    """

    xml_ns = xmlns.fes20

    def get_type_name(self):
        raise NotImplementedError()

    def build_query(self, compiler) -> Q:
        raise NotImplementedError()


@dataclass
@tag_registry.register("ResourceId")
class ResourceId(Id):
    """The ``<fes:ResourceId>`` element.
    This element allow queries to retrieve a resource by their identifier.

    This parses the syntax::

        <fes:ResourceId rid="typename.123" />

    This element is placed inside a :class:`~gisserver.parsers.fes20.filters.Filter`.
    """

    #: A raw "resource identifier". It typically includes the object name,
    #: which is completely unrelated to XML namespacing.
    rid: str

    #: Internal extra attribute, referencing the inferred typename from the :attr:`rid`.
    type_name: str | None

    #: Unused, this is part of additional conformance classes.
    version: int | datetime | VersionActionTokens | NoneType = None
    startTime: datetime | None = None
    endTime: datetime | None = None

    def get_type_name(self):
        """Implemented/override to expose the inferred type name."""
        return self.type_name

    def __post_init__(self):
        if conf.GISSERVER_WFS_STRICT_STANDARD and "." not in self.rid:
            raise ExternalValueError("Expected typename.id format") from None

    @classmethod
    def from_string(cls, rid, ns_aliases: dict[str, str]):
        # Like GeoServer, assume the "name" part of the "resource id" is a QName.
        return cls(
            rid=rid,
            type_name=parse_qname(rid.rpartition(".")[0], ns_aliases),
        )

    @classmethod
    @expect_tag(xmlns.fes20, "ResourceId")
    @expect_no_children
    def from_xml(cls, element):
        version = element.get("version")
        startTime = element.get("startTime")
        endTime = element.get("endTime")

        if version:
            version = auto_cast(version)

        rid = element.get_str_attribute("rid")
        return cls(
            rid=rid,
            type_name=element.parse_qname(rid.rpartition(".")[0]),
            version=version,
            startTime=parse_iso_datetime(startTime) if startTime else None,
            endTime=parse_iso_datetime(endTime) if endTime else None,
        )

    def build_query(self, compiler) -> Q:
        """Render the SQL filter"""
        if self.startTime or self.endTime or self.version:
            raise NotImplementedError(
                "No support for <fes:ResourceId> startTime/endTime/version attributes"
            )

        object_id = self.rid.rpartition(".")[2]

        try:
            # The 'ID' parameter is typed as string, but here we can check
            # whether the database model needs an integer instead.
            compiler.feature_types[0].model._meta.pk.get_prep_value(object_id)
        except (TypeError, ValueError) as e:
            raise InvalidParameterValue(
                f"Invalid resourceId value: {e}", locator="resourceId"
            ) from e

        return Q(pk=object_id)



"""Output rendering logic for GeoJSON."""

from __future__ import annotations

import csv
from datetime import datetime, timezone
from io import StringIO

from django.db import models

from gisserver import conf
from gisserver.db import AsEWKT, get_db_rendered_geometry, replace_queryset_geometries
from gisserver.projection import FeatureProjection, FeatureRelation
from gisserver.types import GeometryXsdElement, XsdElement, XsdTypes

from .base import CollectionOutputRenderer


class CSVRenderer(CollectionOutputRenderer):
    """Fast CSV renderer, using a stream response.

    The complex encoding bits are handled by the "csv" library.
    """

    content_type = "text/csv; charset=utf-8"
    content_disposition = 'attachment; filename="{typenames} {page} {date}.csv"'
    max_page_size = conf.GISSERVER_CSV_MAX_PAGE_SIZE
    chunk_size = 40_000

    #: The outputted CSV dialect. This can be a csv.Dialect subclass
    #: or one of the registered names like: "unix", "excel", "excel-tab"
    dialect = "unix"

    def decorate_queryset(
        self, projection: FeatureProjection, queryset: models.QuerySet
    ) -> models.QuerySet:
        """Make sure relations are included with select-related to avoid N-queries.
        Using prefetch_related() isn't possible with .iterator().
        """
        # First, make sure no array or m2m elements exist,
        # as these are not possible to render in CSV.
        projection.remove_fields(
            lambda e: (e.is_many and not e.is_array)
            or e.type == XsdTypes.gmlCodeType  # gml:name
            or e.type == XsdTypes.gmlBoundingShapeType  # gml:boundedBy
        )

        # All database optimizations
        return super().decorate_queryset(projection, queryset)

    def render_stream(self):
        self.output = output = StringIO()
        writer = csv.writer(output, dialect=self.dialect)

        is_first_collection = True
        for sub_collection in self.collection.results:
            projection = sub_collection.projection
            if is_first_collection:
                is_first_collection = False
            else:
                # Multiple feature types requested, add newlines to separate them
                output.write("\n\n")

            # Write the header
            xsd_elements = projection.xsd_root_elements
            writer.writerow(self.get_header(projection, xsd_elements))

            # Write all rows
            for instance in self.read_features(sub_collection):
                writer.writerow(self.get_row(instance, projection, xsd_elements))

                # Only perform a 'yield' every once in a while,
                # as it goes back-and-forth for writing it to the client.
                if output.tell() > self.chunk_size:
                    csv_chunk = output.getvalue()
                    output.seek(0)
                    output.truncate(0)
                    yield csv_chunk

        yield output.getvalue()

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output."""
        message = super().render_exception(exception)
        buffer = self.output.getvalue()
        return f"{buffer}\n\n{message}\n"

    def get_header(
        self, projection: FeatureProjection, xsd_elements: list[XsdElement], prefix=""
    ) -> list[str]:
        """Return all field names."""
        names = []
        append = names.append
        for xsd_element in xsd_elements:
            if xsd_element.type.is_complex_type:
                names.extend(
                    self.get_header(
                        projection,
                        xsd_elements=projection.xsd_child_nodes[xsd_element],
                        prefix=f"{prefix}{xsd_element.name}.",
                    )
                )
            else:
                append(f"{prefix}{xsd_element.name}")

        return names

    def get_row(
        self, instance: models.Model, projection: FeatureProjection, xsd_elements: list[XsdElement]
    ):
        """Return all field values for a single row."""
        values = []
        append = values.append
        for xsd_element in xsd_elements:
            if xsd_element.type.is_geometry:
                append(self.render_geometry(projection, instance, xsd_element))
                continue

            value = xsd_element.get_value(instance)
            if xsd_element.type.is_complex_type:
                values.extend(
                    self.get_row(
                        instance=value,
                        projection=projection,
                        xsd_elements=projection.xsd_child_nodes[xsd_element],
                    )
                )
            elif isinstance(value, list):
                # Array field
                append(",".join(map(str, value)))
            elif isinstance(value, datetime):
                append(str(value.astimezone(timezone.utc)))
            else:
                append(value)
        return values

    def render_geometry(
        self,
        projection: FeatureProjection,
        instance: models.Model,
        geo_element: GeometryXsdElement,
    ) -> str:
        """Render the contents of a geometry value."""
        geometry = geo_element.get_value(instance)
        projection.output_crs.apply_to(geometry)
        return geometry.ewkt


class DBCSVRenderer(CSVRenderer):
    """Further optimized CSV renderer that uses the database to render EWKT.
    This is about 40% faster than calling the GEOS C-API from python.
    """

    def decorate_queryset(
        self, projection: FeatureProjection, queryset: models.QuerySet
    ) -> models.QuerySet:
        # Instead of reading the binary geometry data, let the database generate EWKT data.
        # As annotations can't be done for select_related() objects, prefetches are used instead.
        queryset = super().decorate_queryset(projection, queryset)
        return replace_queryset_geometries(
            queryset, projection.geometry_elements, projection.output_crs, AsEWKT
        )

    def get_prefetch_queryset(
        self, projection: FeatureProjection, feature_relation: FeatureRelation
    ) -> models.QuerySet | None:
        """Perform DB annotations for prefetched relations too."""
        queryset = super().get_prefetch_queryset(projection, feature_relation)
        if queryset is None:
            return None

        # Find which fields are GML elements, annotate these too.
        return replace_queryset_geometries(
            queryset, feature_relation.geometry_elements, projection.output_crs, AsEWKT
        )

    def render_geometry(
        self,
        projection: FeatureProjection,
        instance: models.Model,
        geo_element: GeometryXsdElement,
    ):
        """Render the geometry using a database-rendered version."""
        return get_db_rendered_geometry(instance, geo_element, AsEWKT)



#!/usr/bin/env python
import re
from pathlib import Path

from setuptools import find_packages, setup


def read(*parts):
    file_path = Path(__file__).parent.joinpath(*parts)
    with open(file_path) as f:
        return f.read()


def find_version(*parts):
    version_file = read(*parts)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return str(version_match.group(1))
    raise RuntimeError("Unable to find version string.")


tests_require = [
    "django-environ >= 0.4.5",
    "psycopg2-binary >= 2.8.4",
    "lxml >= 4.9.1",
    "pytest >= 6.2.3",
    "pytest-django >= 4.1.0",
    "pytest-cov >= 2.11.1",
]

docs_require = [
    "Django ~= 5.1",
    "sphinxcontrib-django >= 2.5",
    "myst-parser >= 3.0.1",
    "psycopg2-binary >= 2.8.4",
]


setup(
    name="django-gisserver",
    version=find_version("gisserver", "__init__.py"),
    license="Mozilla Public License 2.0",
    install_requires=[
        "Django >= 4.2",
        "defusedxml >= 0.7.1",
        "lru_dict >= 1.1.7",
        "orjson >= 3.9.15",
        "pyproj >= 3.6.1",
    ],
    extras_require={
        "tests": tests_require,
        "docs": docs_require,
    },
    requires=["Django (>=4.2, >=5.1)"],
    description="Django speaking WFS 2.0 (exposing GeoDjango model fields)",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Diederik van der Boor",
    author_email="opensource@edoburu.nl",
    url="https://github.com/amsterdam/django-gisserver",
    packages=find_packages(exclude=("tests*", "example*"), include=("gisserver*")),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.1",
        "Framework :: Django :: 5.2",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.10",
)



"""Internal XSD type definitions.

These types are the internal schema definition, and the foundation for all output generation.

The end-users of this library typically create a WFS feature type definition by using
the :class:`~gisserver.features.FeatureType` / :class:`~gisserver.features.FeatureField` classes.

The feature type classes use the model metadata to construct the internal XMLSchema structure.
Nearly all WFS requests are handled by walking this structure (like ``DescribeFeatureType``
or ``GetFeature``). The rendered output is created by walking through this structure,
and writing the XML elements. All search queries (using XPath) are processed by
resolving the corresponding element/attributes, to find to the underlying Django model field.

The structure has the following elements:

* :class:`XsdElement` to define the schema of a single XML element (``<xsd:element>``).
* :class:`XsdAttribute` to define the schema of a single XML attribute (``<xsd:attribute>``).

Each XMLSchema node defines it's "data type" as either:

* :class:`XsdTypes` for simple well-known data types (e.g. text/int, etc..)
* :class:`XsdComplexType` for a complete class definition (holding elements and attributes)

Custom field types could also generate these field types.
"""

from __future__ import annotations

import logging
import operator
import re
from dataclasses import dataclass, field
from datetime import date, datetime, time, timedelta
from decimal import Decimal as D
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Literal

import django
from django.contrib.gis.db.models import F, GeometryField
from django.contrib.gis.geos import GEOSGeometry
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.fields.related import RelatedField

from gisserver.compat import ArrayField, GeneratedField
from gisserver.crs import CRS
from gisserver.exceptions import ExternalParsingError, OperationProcessingFailed
from gisserver.geometries import BoundingBox
from gisserver.parsers import values
from gisserver.parsers.xml import parse_qname, split_ns, xmlns

logger = logging.getLogger(__name__)

__all__ = [
    "GeometryXsdElement",
    "GmlIdAttribute",
    "GmlNameElement",
    "GmlBoundedByElement",
    "ORMPath",
    "XPathMatch",
    "XsdAnyType",
    "XsdAttribute",
    "XsdComplexType",
    "XsdElement",
    "XsdNode",
    "XsdTypes",
]

RE_XPATH_ATTR = re.compile(r"\[[^\]]+\]$")  # match [@attr=..]


class XsdAnyType:
    """Base class for all types used in the XML definition.
    This includes the enum values (:class:`XsdTypes`) for well-known types,
    adn the :class:`XsdComplexType` that represents a while class definition.
    """

    #: Local name of the XML element
    name: str

    #: Namespace of the XML element
    namespace = None

    #: Whether this is a complex type
    is_complex_type = False

    #: Whether this is a geometry
    is_geometry = False  # Overwritten for some gml types.

    def __str__(self):
        """Return the type name (in full XML format)"""
        raise NotImplementedError()

    def to_python(self, raw_value):
        """Convert a raw string value to this type representation"""
        return raw_value


class XsdTypes(XsdAnyType, Enum):
    """Brief enumeration of common XMLSchema types.

    The :class:`XsdElement` and :class:`XsdAttribute` can use these enum members
    to indicate their value is a well-known XML Schema. Some GML types are included as well.

    Each member value is a fully qualified XML name.
    The output rendering will convert these to the chosen prefixes.
    """

    anyType = xmlns.xs.qname("anyType")  # not "xsd:any", that is an element.
    string = xmlns.xs.qname("string")
    boolean = xmlns.xs.qname("boolean")
    decimal = xmlns.xs.qname("decimal")  # the base type for all numbers too.
    integer = xmlns.xs.qname("integer")  # integer value
    float = xmlns.xs.qname("float")
    double = xmlns.xs.qname("double")
    time = xmlns.xs.qname("time")
    date = xmlns.xs.qname("date")
    dateTime = xmlns.xs.qname("dateTime")
    anyURI = xmlns.xs.qname("anyURI")

    # Number variations
    byte = xmlns.xs.qname("byte")  # signed 8-bit integer
    short = xmlns.xs.qname("short")  # signed 16-bit integer
    int = xmlns.xs.qname("int")  # signed 32-bit integer
    long = xmlns.xs.qname("long")  # signed 64-bit integer
    unsignedByte = xmlns.xs.qname("unsignedByte")  # unsigned 8-bit integer
    unsignedShort = xmlns.xs.qname("unsignedShort")  # unsigned 16-bit integer
    unsignedInt = xmlns.xs.qname("unsignedInt")  # unsigned 32-bit integer
    unsignedLong = xmlns.xs.qname("unsignedLong")  # unsigned 64-bit integer

    # Less common, but useful nonetheless:
    duration = xmlns.xs.qname("duration")
    nonNegativeInteger = xmlns.xs.qname("nonNegativeInteger")
    gYear = xmlns.xs.qname("gYear")
    hexBinary = xmlns.xs.qname("hexBinary")
    base64Binary = xmlns.xs.qname("base64Binary")
    token = xmlns.xs.qname("token")  # noqa: S105
    language = xmlns.xs.qname("language")

    # Types that contain a GML value as member:
    # Note these receive the "is_geometry = True" value below.
    gmlGeometryPropertyType = xmlns.gml.qname("GeometryPropertyType")
    gmlPointPropertyType = xmlns.gml.qname("PointPropertyType")
    gmlCurvePropertyType = xmlns.gml.qname("CurvePropertyType")  # curve is base for LineString
    gmlSurfacePropertyType = xmlns.gml.qname("SurfacePropertyType")  # GML2 had PolygonPropertyType
    gmlMultiSurfacePropertyType = xmlns.gml.qname("MultiSurfacePropertyType")
    gmlMultiPointPropertyType = xmlns.gml.qname("MultiPointPropertyType")
    gmlMultiCurvePropertyType = xmlns.gml.qname("MultiCurvePropertyType")
    gmlMultiGeometryPropertyType = xmlns.gml.qname("MultiGeometryPropertyType")

    # Other typical GML values:

    #: The type for ``<gml:name>`` elements.
    gmlCodeType = xmlns.gml.qname("CodeType")  # for <gml:name>

    #: The type for ``<gml:boundedBy>`` elements.
    gmlBoundingShapeType = xmlns.gml.qname("BoundingShapeType")

    #: The type for ``<gml:Envelope>`` elements, sometimes used as function argument type.
    gmlEnvelopeType = xmlns.gml.qname("EnvelopeType")

    #: A direct geometry value, sometimes used as function argument type.
    gmlAbstractGeometryType = xmlns.gml.qname("AbstractGeometryType")

    #: A feature that has a gml:name and gml:boundedBy as possible child element.
    gmlAbstractFeatureType = xmlns.gml.qname("AbstractFeatureType")

    #: The base of gml:AbstractFeatureType
    gmlAbstractGMLType = xmlns.gml.qname("AbstractGMLType")

    def __str__(self):
        return self.value

    def __init__(self, value):
        # Parse XML namespace data once, which to_qname() uses.
        # Can't set enum.name, so will use a property for that.
        self.namespace, self._localname = split_ns(value)
        self.is_geometry = False  # redefined below

    @cached_property
    def name(self) -> str:
        """Overwrites enum.name to return the XML local name.
        This is used for to_qname().
        """
        return self._localname

    @cached_property
    def _to_python_func(self):
        try:
            return TYPES_TO_PYTHON[self]
        except KeyError:
            raise NotImplementedError(f'Casting to "{self}" is not implemented.') from None

    def to_python(self, raw_value):
        """Convert a raw string value to this type representation.

        :raises ExternalParsingError: When the value can't be converted to the proper type.
        """
        if self.is_geometry or isinstance(raw_value, TYPES_AS_PYTHON[self]):
            # Detect when the value was already parsed, no need to reparse a date for example.
            return raw_value

        try:
            return self._to_python_func(raw_value)
        except ExternalParsingError:
            raise  # subclass of ValueError so explicitly caught and reraised
        except (TypeError, ValueError, ArithmeticError) as e:
            # ArithmeticError is base of DecimalException
            logger.debug("Parsing error for %r: %s", raw_value, e)
            name = self.name if self.namespace == xmlns.xsd.value else self.value
            raise ExternalParsingError(f"Can't cast '{raw_value}' to {name}.") from e


for _type in (
    XsdTypes.gmlGeometryPropertyType,
    XsdTypes.gmlPointPropertyType,
    XsdTypes.gmlCurvePropertyType,
    XsdTypes.gmlSurfacePropertyType,
    XsdTypes.gmlMultiSurfacePropertyType,
    XsdTypes.gmlMultiPointPropertyType,
    XsdTypes.gmlMultiCurvePropertyType,
    XsdTypes.gmlMultiGeometryPropertyType,
    # gml:boundedBy is technically a geometry, which we don't support in queries currently.
    XsdTypes.gmlBoundingShapeType,
):
    # One of the reasons the code checks for "xsd_element.type.is_geometry"
    # is because profiling showed that isinstance(xsd_element, ...) is really slow.
    # When rendering 5000 objects with 10+ elements, isinstance() started showing up as hotspot.
    _type.is_geometry = True


def _as_is(v):
    return v


TYPES_AS_PYTHON = {
    XsdTypes.date: date,
    XsdTypes.dateTime: datetime,
    XsdTypes.time: time,
    XsdTypes.string: str,
    XsdTypes.boolean: bool,
    XsdTypes.integer: int,
    XsdTypes.int: int,
    XsdTypes.long: int,
    XsdTypes.short: int,
    XsdTypes.byte: int,
    XsdTypes.unsignedInt: int,
    XsdTypes.unsignedLong: int,
    XsdTypes.unsignedShort: int,
    XsdTypes.unsignedByte: int,
    XsdTypes.float: D,  # auto_cast() always converts to decimal
    XsdTypes.double: D,
    XsdTypes.decimal: D,
    XsdTypes.duration: timedelta,
    XsdTypes.nonNegativeInteger: int,
    XsdTypes.gYear: int,
    XsdTypes.hexBinary: bytes,
    XsdTypes.base64Binary: bytes,
    XsdTypes.token: str,
    XsdTypes.language: str,
    XsdTypes.gmlCodeType: str,
    XsdTypes.anyType: type(Ellipsis),
}

TYPES_TO_PYTHON = {
    **TYPES_AS_PYTHON,
    XsdTypes.date: values.parse_iso_date,
    XsdTypes.dateTime: values.parse_iso_datetime,
    XsdTypes.time: values.parse_iso_time,
    XsdTypes.string: _as_is,
    XsdTypes.boolean: values.parse_bool,
    XsdTypes.duration: values.parse_iso_duration,
    XsdTypes.gmlCodeType: _as_is,
    XsdTypes.anyType: values.auto_cast,
}


class XsdNode:
    """Base class for :class:`XsdElement` and :class:`XsdAttribute`.

    This contains all common mapping/resolving that both elements and attributes share.
    For instance, how XML nodes are mapped into ORM paths, converted into ORM filters,
    parse query input and read model attributes to write as output.
    """

    #: Whether this node is an :class:`XsdAttribute` (avoids slow ``isinstance()`` checks)
    is_attribute = False
    #: Whether this node can occur multiple times.
    is_many = False

    #: The local name of the XML element
    name: str

    #: The data type of the element/attribute, both :class:`XsdComplexType` and :class:`XsdTypes` are allowed.
    type: XsdAnyType

    #: XML Namespace of the element
    namespace: xmlns | str | None

    #: Which field to read from the model to get the value
    #: This supports dot notation to access related attributes.
    source: models.Field | models.ForeignObjectRel | None

    #: Which field to read from the model to get the value
    #: This supports dot notation to access related attributes.
    model_attribute: str | None

    #: A link back to the parent that described the feature this node is a part of.
    #: This helps to perform additional filtering in side meth:get_value: based on user policies.
    feature_type: FeatureType | None

    def __init__(
        self,
        name: str,
        type: XsdAnyType,
        namespace: xmlns | str | None,
        *,
        source: models.Field | models.ForeignObjectRel | None = None,
        model_attribute: str | None = None,
        absolute_model_attribute: str | None = None,
        feature_type: FeatureType | None = None,
    ):
        """
        :param name: The local name of the element.
        :param type: The XML Schema type of the element, can also be a XsdComplexType.
        :param namespace: XML namespace URI.
        :param source: Original Model field, which can provide more metadata/parsing.
        :param model_attribute: The Django model path that this element accesses.
        :param absolute_model_attribute: The full path, including parent elements.
        :param feature_type: Typically assigned in :meth:`~gisserver.features.FeatureField.bind`,
                             needed by some :meth:`get_value` functions.
        """
        if ":" in name:
            raise ValueError(
                "XsdNode should receive the localname, not the QName in ns:localname format."
            )
        elif "}" in name:
            raise ValueError(
                "XsdNode should receive the localname, not the full name in {uri}name format."
            )

        # Using plain assignment instead of dataclass turns out to be needed
        # for flexibility and easier subclassing.
        self.name = name
        self.type = type
        self.namespace = str(namespace) if namespace is not None else None  # cast enum members.
        self.source = source
        self.model_attribute = model_attribute or self.name
        self.absolute_model_attribute = absolute_model_attribute or self.model_attribute
        # link back to top-level parent, some get_value() functions need it.
        self.feature_type = feature_type

        if (
            self.model_attribute
            and self.absolute_model_attribute
            and not self.absolute_model_attribute.endswith(self.model_attribute)
        ):
            raise ValueError("Inconsistent 'absolute_model_attribute' and 'model_attribute' value")

        self._attrgetter = operator.attrgetter(self.model_attribute)
        self._valuegetter = self._build_valuegetter(self.model_attribute, self.source)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: {self.xml_name}"
            f", model_attribute={self.model_attribute}>"
        )

    @staticmethod
    def _build_valuegetter(
        model_attribute: str,
        field: models.Field | models.ForeignObjectRel | None,
    ):
        """Select the most efficient read function to retrieves the value.

        Since reading a value can be called like 150000+ times, this is heavily optimized.

        This attempts to use ``operator.attrgetter()`` whenever possible,
        since this will be much faster than using ``getattr()``.
        The custom ``value_from_object()`` is fully supported too.
        """
        if field is None or isinstance(field, models.ForeignObjectRel):
            # No model field, can only use getattr(). The attrgetter() function is both faster,
            # and has built-in support for traversing model attributes with dots.
            return operator.attrgetter(model_attribute)

        if field.value_from_object.__func__ is models.Field.value_from_object:
            # No custom value_from_object(), this can be fully emulated with attrgetter() too.
            # Still allow the final node to have a custom attname,
            # which is what Field.value_from_object() does.
            names = model_attribute.split(".")
            names[-1] = field.attname
            return operator.attrgetter(".".join(names))

        if "." not in model_attribute:
            # Single level, use the custom value_from_object() directly.
            return field.value_from_object

        # Need to traverse the related field path, and use value_from_object() on the final object.
        names = model_attribute.split(".")
        get_related_instance = operator.attrgetter(".".join(names[:-1]))

        def _related_get_value_from_object(instance):
            related_instance = get_related_instance(instance)
            return field.value_from_object(related_instance)

        return _related_get_value_from_object

    @cached_property
    def is_array(self) -> bool:
        """Tell whether this node is backed by an PostgreSQL Array Field."""
        return ArrayField is not None and isinstance(self.source, ArrayField)

    @cached_property
    def is_flattened(self) -> bool:
        """Whether the field is a lookup to a relation."""
        return "." in self.model_attribute

    @cached_property
    def xml_name(self):
        """The XML element/attribute name."""
        return f"{{{self.namespace}}}{self.name}" if self.namespace else self.name

    def relative_orm_path(self, parent: XsdElement | None = None) -> str:
        """The ORM field lookup to perform, relative to the parent element."""
        if parent is None:
            return self.orm_path

        prefix = f"{parent.orm_path}__"
        if not self.orm_path.startswith(prefix):
            raise ValueError(f"Node '{self}' is not a child of '{parent}.")
        else:
            return self.orm_path[len(prefix) :]

    @cached_property
    def local_orm_path(self) -> str:
        """The ORM field lookup to perform."""
        if self.model_attribute is None:
            raise ValueError(f"Node {self.xml_name} has no 'model_attribute' set.")
        return self.model_attribute.replace(".", "__")

    @cached_property
    def orm_path(self) -> str:
        """The ORM field lookup to perform."""
        if self.absolute_model_attribute is None:
            raise ValueError(f"Node {self.xml_name} has no 'absolute_model_attribute' set.")
        return self.absolute_model_attribute.replace(".", "__")

    @cached_property
    def orm_field(self) -> str:
        """The direct ORM field that provides this property; the first relative level.
        Typically, this is the same as the field name.
        """
        return self.orm_path.partition(".")[0]

    @cached_property
    def orm_relation(self) -> tuple[str | None, str]:
        """The ORM field and parent relation.
        Note this isn't something like "self.parent.orm_path",
        as this mode may have a dotted-path to its source attribute.
        """
        path, _, field = self.orm_path.rpartition("__")
        return path or None, field

    def build_lhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element is used as left-hand-side of a comparison.
        This is needed for queries like "<element> == <value>"
        """
        return match.orm_path

    def build_rhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element would be used as right-hand-side.
        This is needed for queries like "<value> == <element>" or "<element> == <element>".
        """
        return F(match.orm_path)

    def get_value(self, instance: models.Model):
        """Provide the value for the data"""
        # For foreign keys, it's not possible to use the model value,
        # as that would conflict with the field type in the XSD schema.
        try:
            if self.type.is_complex_type:
                # This element has sub elements, which need the Django model instance.
                # Avoid unwanted value_from_object(), instead return the model instance.
                value = self._attrgetter(instance)
                if self.is_many and isinstance(value, models.Manager):
                    # Make sure callers can read the individual items by iterating over the value.
                    value = value.all()
                    if self.feature_type is not None:
                        return self.feature_type.filter_related_queryset(value)
                return value
            else:
                # the _valuegetter() supports value_from_object() on custom fields.
                return self._valuegetter(instance)
        except (AttributeError, ObjectDoesNotExist):
            # E.g. Django foreign keys that point to a non-existing member.
            return None

    def format_raw_value(self, value):
        """Allow to apply some final transformations on a value.
        This is mainly used to support @gml:id which includes a prefix.
        """
        return value

    def to_python(self, raw_value: str):
        """Convert a raw value to the Python data type for this element type.
        :raises ValidationError: When the value isn't allowed for the field type.
        """
        try:
            raw_value = self.type.to_python(raw_value)
            if self.source is not None:
                raw_value = self.source.get_prep_value(raw_value)
        except ValidationError as e:
            raise ValidationError(
                f"Invalid data for the '{self.name}' property: {e.messages[0]}",
                code=e.code,
            ) from e
        except (ValueError, TypeError) as e:
            raise ValidationError(f"Invalid data for the '{self.name}' property: {e}") from e

        return raw_value

    def validate_comparison(self, raw_value: str, lookup, tag=None):
        """Validate whether the input value can be used in a comparison.
        This avoids comparing a database DATETIME object to an integer.

        The raw string value can be passed here. Auto-cased values could
        raise an TypeError due to being unsupported by the validation.

        :param raw_value: The string value taken from the XML node.
        :param lookup: The ORM lookup (e.g. ``equals`` or ``fes_like``).
        :param tag: The filter operator tag name, e.g. ``PropertyIsEqualTo``.
        :returns: The parsed Python value.
        """
        # Not calling self.source.validate() as that checks for allowed choices,
        # which shouldn't be checked against for a filter query.
        raw_value = self.to_python(raw_value)

        # Check whether the Django model field supports the lookup
        # This prevents calling LIKE on a datetime or float field.
        # For foreign keys, this depends on the target field type.
        if (
            self.source is not None
            and self.source.get_lookup(lookup) is None
            or (
                isinstance(self.source, RelatedField)
                and self.source.target_field.get_lookup(lookup) is None
            )
        ):
            logger.debug(
                "Model field '%s.%s' does not support ORM lookup '%s' used by '%s'.",
                self.feature_type.model._meta.model_name,
                self.absolute_model_attribute,
                lookup,
                tag,
            )
            raise OperationProcessingFailed(
                f"Operator '{tag}' is not supported for the '{self.name}' property.",
                locator="filter",
                status_code=400,  # not HTTP 500 here. Spec allows both.
            )

        return raw_value


class XsdElement(XsdNode):
    """Declare an XSD element.

    Typically, this maps into a Django model field.

    This holds the definition for a single property in the WFS server.
    It's used in ``DescribeFeatureType`` to output the field metadata,
    and used in ``GetFeature`` to access the actual value from the object.
    Overriding :meth:`XsdNode.get_value` allows to override this logic.

    The :attr:`name` may differ from the underlying :attr:`XsdNode.model_attribute`,
    so the WFS server can use other field names then the underlying model.

    A dotted-path notation can be used for :attr:`XsdNode.model_attribute` to access
    a related field. For the WFS client, the data appears to be flattened.
    """

    #: Whether the element can be null
    nillable: bool | None
    #: The minimal number of times the element occurs in the output.
    min_occurs: int | None
    #: The maximum number of times this element occurs in the output.
    max_occurs: int | Literal["unbounded"] | None

    def __init__(
        self,
        name: str,
        type: XsdAnyType,
        namespace: xmlns | str | None,
        *,
        nillable: bool | None = None,
        min_occurs: int | None = None,
        max_occurs: int | Literal["unbounded"] | None = None,
        source: models.Field | models.ForeignObjectRel | None = None,
        model_attribute: str | None = None,
        absolute_model_attribute: str | None = None,
        feature_type: FeatureType | None = None,
    ):
        super().__init__(
            name,
            type,
            namespace=namespace,
            source=source,
            model_attribute=model_attribute,
            absolute_model_attribute=absolute_model_attribute,
            feature_type=feature_type,
        )
        self.nillable = nillable
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs

    @cached_property
    def is_many(self) -> bool:
        """Tell whether the XML element can be rendered multiple times.
        Note this happens both with array fields and "..._to_many" relations.
        """
        return self.max_occurs and (
            self.max_occurs == "unbounded"
            or self.max_occurs > 1
            # needed for ArrayField(size=1):
            or (ArrayField is not None and isinstance(self.source, ArrayField))
        )


class _XsdElement_WithComplexType(XsdElement):
    """This only exists as "protocol" for the type annotations"""

    type: XsdComplexType


class XsdAttribute(XsdNode):
    """Declare an XSD attribute.

    Typically, this maps into a Django model field.

    Most fields are mapped into XML Elements (:class:`XsdElement`).
    However, WFS also supports XML attributes, and queries against them.
    This class is uses to support filters against attributes like "gml:id".
    """

    is_attribute = True

    type: XsdTypes
    use: str = "optional"

    def __init__(
        self,
        name: str,
        type: XsdAnyType = XsdTypes.string,  # added default
        *,
        namespace: xmlns | str | None = None,
        use: str = "optional",
        source: models.Field | models.ForeignObjectRel | None = None,
        model_attribute: str | None = None,
        absolute_model_attribute: str | None = None,
        feature_type: FeatureType | None = None,
    ):
        super().__init__(
            name,
            type,
            namespace=namespace,
            source=source,
            model_attribute=model_attribute,
            absolute_model_attribute=absolute_model_attribute,
            feature_type=feature_type,
        )
        self.use = use


class GeometryXsdElement(XsdElement):
    """A subtype for the :class:`XsdElement` that provides access to geometry data.

    This declares an element such as::

        <app:geometry>
            <gml:Point>...</gml:Point>
        </app:geometry>

    Hence, the :attr:`namespace` of this element isn't the GML namespace,
    only the type it points to is geometry data.

    The :attr:`source` is guaranteed to point to a :class:`~django.contrib.gis.models.GeometryField`,
    and can be a :class:`~django.db.models.GeneratedField` in Django 5
    as long as its ``output_field`` points to a :class:`~django.contrib.gis.models.GeometryField`.
    """

    if django.VERSION >= (5, 0):
        source: GeometryField | models.GeneratedField
    else:
        source: GeometryField

    @cached_property
    def source_srid(self) -> int:
        """Tell which Spatial Reference Identifier the source information is stored under."""
        if GeneratedField is not None and isinstance(self.source, GeneratedField):
            # Allow GeometryField to be wrapped as:
            # models.GeneratedField(SomeFunction("geofield"), output_field=models.GeometryField())
            return self.source.output_field.srid
        else:
            return self.source.srid


class GmlIdAttribute(XsdAttribute):
    """A virtual ``gml:id="..."`` attribute that can be queried.
    This subclass has overwritten :meth:`get_value` logic to format the value.
    """

    type_name: str

    def __init__(
        self,
        type_name: str,
        source: models.Field | models.ForeignObjectRel | None = None,
        model_attribute="pk",
        absolute_model_attribute=None,
        feature_type: FeatureType | None = None,
    ):
        super().__init__(
            name="id",
            namespace=xmlns.gml,
            source=source,
            model_attribute=model_attribute,
            absolute_model_attribute=absolute_model_attribute,
            feature_type=feature_type,
        )
        object.__setattr__(self, "type_name", type_name)

    def get_value(self, instance: models.Model):
        """Render the value."""
        pk = super().get_value(instance)  # handle dotted-name notations
        return f"{self.type_name}.{pk}"

    def format_raw_value(self, value):
        """Format the value as retrieved from the database."""
        return f"{self.type_name}.{value}"


class GmlNameElement(XsdElement):
    """A subclass to handle the ``<gml:name>`` element.
    This displays a human-readable title for the object.

    Currently, this just reads a single attribute,
    but it can be extended to support formatted names
    (although that would make comparisons on ``element@gml:name`` more complex).
    """

    def __init__(
        self,
        model_attribute: str,
        source: models.Field | models.ForeignObjectRel | None = None,
        feature_type=None,
    ):
        # Prefill most known fields
        super().__init__(
            name="name",
            type=XsdTypes.gmlCodeType,
            namespace=xmlns.gml,
            min_occurs=0,
            source=source,
            model_attribute=model_attribute,
            feature_type=feature_type,
        )

    def get_value(self, instance: models.Model):
        """Override value retrieval to retrieve the value from the feature type."""
        if self.feature_type is not None:
            # Let FeatureType provide a nice display/title for the object.
            return self.feature_type.get_display_value(instance)
        else:
            # Fallback, when using this class at a sub-level object.
            return super().get_value(instance)


class GmlBoundedByElement(XsdElement):
    """A subclass to handle the ``<gml:boundedBy>`` element.

    This override makes sure this non-model element data
    can be included in the XML tree like every other element.
    Its value is the complete bounding box of the feature type data.
    """

    def __init__(self, feature_type):
        # Prefill most known fields
        super().__init__(
            name="boundedBy",
            type=XsdTypes.gmlBoundingShapeType,
            namespace=xmlns.gml,
            min_occurs=0,
            feature_type=feature_type,
        )
        self.model_attribute = None

    def build_lhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element is used as
        left-hand-side of a comparison."""
        return compiler.add_annotation(self.build_rhs_part(compiler, match))

    def build_rhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element would be used as right-hand-side"""
        raise NotImplementedError("queries against <gml:boundedBy> are not supported")

    def get_value(self, instance: models.Model, crs: CRS | None = None) -> BoundingBox | None:
        """Provide the value of the <gml:boundedBy> field,
        which is the bounding box for a single instance.

        This is only used for native Python rendering. When the database
        rendering is enabled (GISSERVER_USE_DB_RENDERING=True), the calculation
        is entirely performed within the query.
        """
        geometries: list[GEOSGeometry] = list(
            # remove 'None' values
            filter(
                None,
                [
                    # support dotted paths here for geometries in a foreign key relation.
                    operator.attrgetter(geo_element.absolute_model_attribute)(instance)
                    for geo_element in self.feature_type.all_geometry_elements
                ],
            )
        )
        if not geometries:
            return None

        return BoundingBox.from_geometries(geometries, crs)


@dataclass(frozen=True)
class XsdComplexType(XsdAnyType):
    """Define an ``<xsd:complexType>`` that represents a whole class definition.

    Typically, this maps into a Django model, with each element pointing to a model field.
    For example:

    .. code-block:: python

        XsdComplexType(
            "PersonType",
            elements=[
                XsdElement("name", type=XsdTypes.string),
                XsdElement("age", type=XsdTypes.integer),
                XsdElement("address", type=XsdComplexType(
                    "AddressType",
                    elements=[
                        XsdElement("street", type=XsdTypes.string),
                        ...
                    ]
                )),
            ],
            attributes=[
                XsdAttribute("id", type=XsdTypes.integer),
            ],
        )

    A complex type can hold multiple :class:`XsdElement` and :class:`XsdAttribute`
    nodes as children, composing an object. Its :attr:`base` may point to a :class:`XsdComplexType`
    as base class, allowing to define those inherited elements too.

    Each element can be a complex type themselves, to create a nested class structure.
    That also allows embedding models with their relations into a single response.

    .. note:: Good to know
        This object definition is the internal "source of truth" regarding
        which field names and field elements are used in the WFS server:

        * The ``DescribeFeatureType`` request uses this definition to render the matching XMLSchema.
        * Incoming XPath queries are parsed using this object to resolve the XPath to model attributes.

    Objects of this type are typically generated by the :class:`~gisserver.features.FeatureType` and
    :class:`~gisserver.features.ComplexFeatureField` classes, using the Django model data.

    By default, The :attr:`base` type is detected as ``<gml:AbstractFeatureType>``,
    when there is a geometry element in the definition.
    """

    #: Internal class name (without XML namespace/prefix)
    name: str

    #: The XML namespace
    namespace: str | None

    #: All local elements in this class
    elements: list[XsdElement] = field(default_factory=list)

    #: All attributes in this class
    attributes: list[XsdAttribute] = field(default_factory=list)

    #: The base class of this type. Typically gml:AbstractFeatureType,
    #: which provides the <gml:name> and <gml:boundedBy> elements.
    base: XsdAnyType | None = None

    #: The Django model class that this type was based on.
    source: type[models.Model] | None = None

    def __post_init__(self):
        # Autodetect (or autocorrect) to have the proper base class when gml elements are present.
        if self.base is None:
            if any(e.type.is_geometry for e in self.elements):
                # for <gml:name> and <gml:boundedBy> elements.
                self.__dict__["base"] = XsdTypes.gmlAbstractFeatureType
            elif any(e.type is XsdTypes.gmlCodeType for e in self.elements):
                # for <gml:name> only
                self.__dict__["base"] = XsdTypes.gmlAbstractGMLType

    def __str__(self):
        return self.xml_name

    @cached_property
    def xml_name(self):
        """Name in the XMLSchema (e.g. {http://example.org/namespace}:SomeClass)."""
        return f"{{{self.namespace}}}{self.name}" if self.namespace else self.name

    @property
    def is_complex_type(self) -> bool:
        """Always indicates this is a complex type."""
        return True  # a property to avoid being used as field.

    @cached_property
    def elements_including_base(self) -> list[XsdElement]:
        """The local and inherited elements of this XSD type."""
        if self.base is not None and self.base.is_complex_type:
            # Add all base class members, in their correct ordering
            # By having these as XsdElement objects instead of hard-coded writes,
            # the query/filter logic also works for these elements.
            return self.base.elements + self.elements
        else:
            return self.elements

    @cached_property
    def geometry_elements(self) -> list[GeometryXsdElement]:
        """Shortcut to get all geometry elements"""
        return [e for e in self.elements if e.type.is_geometry]

    @cached_property
    def complex_elements(self) -> list[_XsdElement_WithComplexType]:
        """Shortcut to get all elements with a complex type.
        To get all complex elements recursively, read :attr:`all_complex_elements`.
        """
        return [e for e in self.elements if e.type.is_complex_type]

    @cached_property
    def flattened_elements(self) -> list[XsdElement]:
        """Shortcut to get all elements with a flattened model attribute"""
        return [e for e in self.elements if e.is_flattened]

    @cached_property
    def all_complex_elements(self) -> list[_XsdElement_WithComplexType]:
        """Shortcut to get all elements with children.
        This mainly exists to provide a structure to mimic what's used
        when PROPERTYNAME is part of the request.
        """
        child_nodes = []
        for xsd_element in self.complex_elements:
            child_nodes.append(xsd_element)
            child_nodes.extend(xsd_element.type.all_complex_elements)
        return child_nodes

    def resolve_element_path(self, xpath: str, ns_aliases: dict[str, str]) -> list[XsdNode] | None:
        """Resolve a xpath reference to the actual node.
        This returns the whole path, including in-between relations, if a match was found.

        This is used by :meth:`~gisserver.features.FeatureType.resolve_element`
        to convert a request XPath element into the ORM attributes for database queries.
        """
        try:
            pos = xpath.index("/")
            node_name = xpath[:pos]
        except ValueError:
            node_name = xpath
            pos = 0

        # Strip any [@attr=..] conditions
        node_name = RE_XPATH_ATTR.sub("", node_name)

        if node_name.startswith("@"):
            # Resolve attributes (e.g. gml:id)
            if pos:
                return None  # invalid attribute

            xml_name = node_name[1:]
            attribute = self._find_attribute(xml_name=parse_qname(xml_name, ns_aliases))
            return [attribute] if attribute is not None else None
        else:
            element = self._find_element(xml_name=parse_qname(node_name, ns_aliases))
            if element is None:
                return None

            if pos:
                if not element.type.is_complex_type:
                    return None
                else:
                    # Recurse into the child node to find the next part
                    child_path = element.type.resolve_element_path(xpath[pos + 1 :], ns_aliases)
                    return [element] + child_path if child_path is not None else None
            else:
                return [element]

    def _find_element(self, xml_name: str) -> XsdElement | None:
        """Locate an element by name"""
        for element in self.elements:
            if element.xml_name == xml_name:
                return element

        # When there is a base class, resolve elements there too.
        if self.base is not None and self.base.is_complex_type:
            return self.base._find_element(xml_name)
        return None

    def _find_attribute(self, xml_name: str) -> XsdAttribute | None:
        """Locate an attribute by name"""
        for attribute in self.attributes:
            if attribute.xml_name == xml_name:
                return attribute

        # When there is a base class, resolve attributes there too.
        if self.base is not None and self.base.is_complex_type:
            return self.base._find_attribute(xml_name)
        return None


class ORMPath:
    """Base class to provide raw XPath results.

    This base class is designed to allow other query types (besides XPath) too,
    and allows inserting raw data directly to the query compiler (for unit testing).
    """

    def __init__(self, orm_path: str, orm_filters: Q | None = None, is_many=False):
        """Base constructor just assigns items.
        Overwritten classes likely replace this with properties.
        """
        self.orm_path = orm_path
        self.orm_filters = orm_filters
        self.is_many = is_many

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(orm_path={self.orm_path!r}"
            f", orm_filters={self.orm_filters!r})"
        )

    def build_lhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element is used as left-hand-side of a comparison.
        For example: ``path == value``.
        """
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.orm_path

    def build_rhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element would be used as right-hand-side.
        For example: ``path1 == path2`` or ``value == path``.
        """
        if self.is_many:
            compiler.add_distinct()
        return F(self.build_lhs(compiler))


class XPathMatch(ORMPath):
    """The ORM path result from am XPath query.

    This result object defines how to resolve an XPath to an ORM object.
    """

    #: The matched element, with all it's parents.
    nodes: list[XsdNode]

    #: The source XPath query
    query: str

    #: The additional filters are needed (due to [@attr=..] syntax).
    orm_filters: Q | None

    def __init__(self, feature_type: FeatureType, nodes: list[XsdNode], query: str):
        self.feature_type = feature_type
        self.nodes = nodes
        self.query = query
        self.orm_filters = None

        if "[" in self.query:
            # If there is an element[@attr=...]/field tag,
            # the build_...() logic should return a Q() object.
            raise NotImplementedError(f"Complex XPath queries are not supported yet: {self.query}")

    @property
    def orm_path(self) -> str:
        """Give the Django ORM path (field__relation__relation2) to the result."""
        return self.nodes[-1].orm_path

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, item) -> XsdNode:
        return self.nodes[item]

    def __repr__(self):
        return f"XPathMatch(nodes={self.nodes!r}, query={self.query!r})"

    @property
    def child(self) -> XsdNode:
        """Return only the final element"""
        return self.nodes[-1]

    @cached_property
    def is_many(self) -> bool:
        """Return whether this ORM path walks over an element that occurs multiple times"""
        return any(node.is_many for node in self.nodes)

    def build_lhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element is used as left-hand-side of a comparison.
        For example: ``path == value``.
        """
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.child.build_lhs_part(compiler, self)

    def build_rhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element would be used as right-hand-side.
        For example: ``path1 == path2`` or ``value == path``.
        """
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.child.build_rhs_part(compiler, self)


if TYPE_CHECKING:
    from .features import FeatureType
    from .parsers.query import CompiledQuery
